/*
 * Copyright (c) 2020 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.android.ccid;

import android.hardware.usb.UsbDevice;
import android.hardware.usb.UsbManager;
import android.util.Log;

import java.io.IOException;

import opencard.core.terminal.CardID;
import opencard.core.terminal.CardTerminal;
import opencard.core.terminal.CardTerminalException;
import opencard.core.terminal.CardTerminalRegistry;
import opencard.core.terminal.CommandAPDU;
import opencard.core.terminal.Pollable;
import opencard.core.terminal.ResponseAPDU;

public class CcidCardTerminal extends CardTerminal implements Pollable {

    CcidUSB reader = null;
    byte[] atr = null;
    boolean cardPresent = false;


    public CcidCardTerminal(String name, UsbDevice usbdevice, UsbManager usbmanager)
            throws CardTerminalException {
        super(name, "CCID", "");

        addSlots(1);

        // Add pollable before creating reader, as addPollable will
        // call poll() on this thread
        CardTerminalRegistry.getRegistry().addPollable(this);
        reader = CcidReader.createCcidReader(usbdevice, usbmanager);
    }



    @Override
    public CardID getCardID(int i) throws CardTerminalException {
        return new CardID(this, 0, atr);
    }



    @Override
    public boolean isCardPresent(int i) throws CardTerminalException {
        if (reader == null)
            return false;

        try {
            synchronized(reader) {
                int status = reader.PC_to_RDR_GetSlotStatus();
                return status == 0 || status == 1;
            }
        } catch (IOException ioe) {
            return false;
        }

    }



    @Override
    public void open() throws CardTerminalException {
        if (reader == null) {
            throw new CardTerminalException("CCID reader not available");
        }

        try {
            synchronized(reader) {
                atr = reader.PC_to_RDR_IccPowerOn();
            }
        } catch (IOException ioe) {
            throw new CardTerminalException(ioe.getLocalizedMessage());
        }
    }



    @Override
    public void close() throws CardTerminalException {
        if (reader == null) {
            return;
        }

        try {
            synchronized(reader) {
                reader.PC_to_RDR_IccPowerOff();
            }
            reader = null;
        } catch (IOException ioe) {
            throw new CardTerminalException(ioe.getLocalizedMessage());
        }
    }



    @Override
    protected CardID internalReset(int i, int i1) throws CardTerminalException {
        if (reader == null) {
            throw new CardTerminalException("CCID reader not available");
        }
        try {
            synchronized(reader) {
                reader.PC_to_RDR_IccPowerOff();
                atr = reader.PC_to_RDR_IccPowerOn();
            }
        } catch (IOException ioe) {
            throw new CardTerminalException(ioe.getLocalizedMessage());
        }
        return new CardID(this, 0, atr);
    }



    @Override
    protected ResponseAPDU internalSendAPDU(int i, CommandAPDU commandAPDU, int i1) throws CardTerminalException {
        if (reader == null) {
            throw new CardTerminalException("CCID reader not available");
        }

        byte[] rsp = null;

        try {
            synchronized(reader) {
                rsp = reader.ccidAPDUProcess(commandAPDU.getBytes());
            }
        } catch (IOException ioe) {
            throw new CardTerminalException(ioe.getLocalizedMessage());
        }

        return new ResponseAPDU(rsp);
    }



    @Override
    public void poll() throws CardTerminalException {
        if (reader == null) {
            return;
        }

        boolean newStatus = isCardPresent(0x00);

        if (newStatus != cardPresent) {
            cardPresent = newStatus;
            if (cardPresent) {
                internalReset(0, 0);
                cardInserted(0);
            } else {
                cardRemoved(0);
            }
        }
    }



    public void detach() {
        reader = null;
        cardRemoved(0);
    }
}
