/*
 * Copyright (c) 2020 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.android.ccid;

import android.app.PendingIntent;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.hardware.usb.UsbDevice;
import android.hardware.usb.UsbManager;
import android.util.Log;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.Iterator;

import opencard.core.terminal.CardTerminalException;
import opencard.core.terminal.CardTerminalFactory;
import opencard.core.terminal.CardTerminalRegistry;
import opencard.core.terminal.TerminalInitException;

public class CcidCardTerminalFactory extends BroadcastReceiver implements CardTerminalFactory {

    private final static Logger logger = LoggerFactory.getLogger(CcidCardTerminalFactory.class);

    public final static String ACTION_USB_PERMISSION = "de.cardcontact.USB_PERMISSION";



    @Override
    public void createCardTerminals(CardTerminalRegistry ctr, String[] strings) throws CardTerminalException, TerminalInitException {
        throw new TerminalInitException("Can not use default createCardTermninal(). Use variant with UsbManager instead");
    }



    public void createCardTerminals(CardTerminalRegistry ctr, Context context) throws CardTerminalException, TerminalInitException {
        UsbManager manager = context.getSystemService(UsbManager.class);

        HashMap<String, UsbDevice> map = manager.getDeviceList();

        Iterator<String> it = map.keySet().iterator();
        while (it.hasNext()) {
            UsbDevice device = map.get(it.next());

            if (!manager.hasPermission(device)) {
                PendingIntent permissionIntent = PendingIntent.getBroadcast(context, 0, new Intent(CcidCardTerminalFactory.ACTION_USB_PERMISSION), 0);
                manager.requestPermission(device, permissionIntent);
            } else {
                String name = device.getProductName() + "(" + device.getSerialNumber() + ")";
                try {
                    CcidCardTerminal terminal = new CcidCardTerminal(name, device, manager);
                    ctr.add(terminal);
                } catch (Exception e) {
                    logger.error("Could not create card terminal", e);
                }
            }
        }
    }



    @Override
    public void open() throws CardTerminalException {

    }



    @Override
    public void close() throws CardTerminalException {

    }



    @Override
    public void onReceive(Context context, Intent intent) {
        CardTerminalRegistry ctr = CardTerminalRegistry.getRegistry();
        UsbManager manager = context.getSystemService(UsbManager.class);
        UsbDevice device = (UsbDevice)intent.getParcelableExtra(UsbManager.EXTRA_DEVICE);

        if (!USBDeviceHandler.isSupported(device)) {
            return;
        }

        if (UsbManager.ACTION_USB_DEVICE_ATTACHED.equals(intent.getAction()) || CcidCardTerminalFactory.ACTION_USB_PERMISSION.equals(intent.getAction())) {
            if (!manager.hasPermission(device)) {
                logger.debug("USB device permission requested");
                PendingIntent permissionIntent = PendingIntent.getBroadcast(context, 0, new Intent(CcidCardTerminalFactory.ACTION_USB_PERMISSION), 0);
                manager.requestPermission(device, permissionIntent);
            } else {
                String name = device.getProductName() + "(" + device.getSerialNumber() + ")";
                logger.debug("USB device " + name + " attached");
                try {
                    CcidCardTerminal terminal = new CcidCardTerminal(name, device, manager);
                    ctr.add(terminal);
                } catch (Exception e) {
                    logger.error("Could not create card terminal", e);
                }
            }
        } else if (UsbManager.ACTION_USB_DEVICE_DETACHED.equals(intent.getAction())) {
            String name = device.getProductName() + "(" + device.getSerialNumber() + ")";
            logger.debug("USB device " + name + " detached");
            CcidCardTerminal ct = (CcidCardTerminal)ctr.cardTerminalForName(name);
            if (ct != null) {
                ctr.removePollable(ct);
                ct.detach();
                try {
                    ctr.remove(ct);
                }
                catch(CardTerminalException cte) {
                    logger.debug("Ignoring ", cte);
                }
            }
        } else {
            logger.debug("USB event " + intent.getAction());
        }
    }
}
