/*
 * Copyright (c) 2020 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.android.ccid;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.EnumSet;
import java.util.LinkedList;
import java.util.List;

public class CcidDescriptor {
    public static enum Voltage {
        _5_0V,
        _3_0V,
        _1_8V
    }

    public static enum Protocol {
        T0,
        T1
    }

    public static enum Mechanical {
        Accept,
        Ejection,
        Capture,
        LockUnlock
    }

    public static enum Feature {
        AutoParamConfigViaATR,
        AutoActivationOnInsert,
        AutoVoltageSelection,
        AutoClockChange,
        AutoDataRateChange,
        AutoParamNego,
        AutoPPS,
        CanStopClock,
        NADAccepted,
        AutoIFSDExchange,
        TPDU,
        ShortAPDU,
        ShortAndExtendedAPDU,
        WakeOnCardAction
    }

    public static enum PINSupport {
        Verification,
        Modification
    }

    private int ccidVersion;
    private int maxSlotIndex;
    private EnumSet<Voltage> voltages;
    private EnumSet<Protocol> protocols;
    private int defaultClock;
    private int maxClock;
    private int numClockSupported;
    private int defaultDataRate;
    private int maxDataRate;
    private int numDataRatesSupported;
    private int maxIFSD;
    private EnumSet<Mechanical> mechanicals;
    private EnumSet<Feature> features;
    private int maxCCIDMessageLength;
    private int classGetResponse;
    private int classEnvelope;
    private EnumSet<PINSupport> pinSupports;
    private int maxCCIDBusySlots;

    public static CcidDescriptor parse(byte[] data) throws IOException {
        ByteBuffer bb = ByteBuffer.wrap(data);
        bb = bb.order(ByteOrder.LITTLE_ENDIAN);

        CcidDescriptor ccid = new CcidDescriptor();
        while (bb.hasRemaining()) {

            int len = bb.get();
            int type = bb.get();
            if (type == 0x04) {
                if (len != 0x09)
                    throw new IllegalArgumentException("Invalid Interface descriptor (wrong length)");
                short i = bb.getShort();
                bb.position(bb.position() + len - 4);
            } else if (type == 0x21) {
                if (len != 0x36)
                    throw new IllegalArgumentException("Invalid Smart Card Device descriptor (wrong length)");

                ccid.ccidVersion = bb.getShort();
                ccid.maxSlotIndex = bb.get();

                int vsData = bb.get();
                List<Voltage> vsList = new LinkedList<Voltage>();
                if ((vsData & 0x01) == 0x01) vsList.add(Voltage._5_0V);
                if ((vsData & 0x02) == 0x02) vsList.add(Voltage._3_0V);
                if ((vsData & 0x04) == 0x04) vsList.add(Voltage._1_8V);
                ccid.voltages = vsList.isEmpty() ? EnumSet.noneOf(Voltage.class) : EnumSet.copyOf(vsList);

                int pData = bb.getInt();
                List<Protocol> pList = new LinkedList<Protocol>();
                if ((pData & 0x00000001) == 0x00000001) pList.add(Protocol.T0);
                if ((pData & 0x00000002) == 0x00000002) pList.add(Protocol.T1);
                ccid.protocols = pList.isEmpty() ? EnumSet.noneOf(Protocol.class) : EnumSet.copyOf(pList);

                ccid.defaultClock = bb.getInt();
                ccid.maxClock = bb.getInt();
                ccid.numClockSupported = bb.get();
                ccid.defaultDataRate = bb.getInt();
                ccid.maxDataRate = bb.getInt();
                ccid.numDataRatesSupported = bb.get();
                ccid.maxIFSD = bb.getInt();
                bb.getInt(); //We ignore the synchProtocols (not relevant for USB)

                int mData = bb.getInt();
                List<Mechanical> mList = new LinkedList<Mechanical>();
                if ((mData & 0x00000001) == 0x00000001) mList.add(Mechanical.Accept);
                if ((mData & 0x00000002) == 0x00000002) mList.add(Mechanical.Ejection);
                if ((mData & 0x00000004) == 0x00000004) mList.add(Mechanical.Capture);
                if ((mData & 0x00000008) == 0x00000008) mList.add(Mechanical.LockUnlock);
                ccid.mechanicals = mList.isEmpty() ? EnumSet.noneOf(Mechanical.class) : EnumSet.copyOf(mList);

                int fData = bb.getInt();
                List<Feature> fList = new LinkedList<Feature>();
                if ((fData & 0x00000002) == 0x00000002)
                    fList.add(Feature.AutoParamConfigViaATR);
                if ((fData & 0x00000004) == 0x00000004)
                    fList.add(Feature.AutoActivationOnInsert);
                if ((fData & 0x00000008) == 0x00000008) fList.add(Feature.AutoVoltageSelection);
                if ((fData & 0x00000010) == 0x00000010) fList.add(Feature.AutoClockChange);
                if ((fData & 0x00000020) == 0x00000020) fList.add(Feature.AutoDataRateChange);
                if ((fData & 0x00000040) == 0x00000040) fList.add(Feature.AutoParamNego);
                if ((fData & 0x00000080) == 0x00000080) fList.add(Feature.AutoPPS);
                if ((fData & 0x00000100) == 0x00000100) fList.add(Feature.CanStopClock);
                if ((fData & 0x00000200) == 0x00000200) fList.add(Feature.NADAccepted);
                if ((fData & 0x00000400) == 0x00000400) fList.add(Feature.AutoIFSDExchange);
                if ((fData & 0x00010000) == 0x00010000) fList.add(Feature.TPDU);
                if ((fData & 0x00020000) == 0x00020000) fList.add(Feature.ShortAPDU);
                if ((fData & 0x00040000) == 0x00040000) fList.add(Feature.ShortAndExtendedAPDU);
                if ((fData & 0x00100000) == 0x00100000) fList.add(Feature.WakeOnCardAction);
                ccid.features = fList.isEmpty() ? EnumSet.noneOf(Feature.class) : EnumSet.copyOf(fList);

                ccid.maxCCIDMessageLength = bb.getInt();
                ccid.classGetResponse = bb.get();
                ccid.classEnvelope = bb.get();

                int psData = bb.get();
                List<PINSupport> psList = new LinkedList<PINSupport>();
                if ((psData & 0x01) == 0x01) psList.add(PINSupport.Verification);
                if ((psData & 0x02) == 0x02) psList.add(PINSupport.Modification);
                ccid.pinSupports = psList.isEmpty() ? EnumSet.noneOf(PINSupport.class) : EnumSet.copyOf(psList);

                ccid.maxCCIDBusySlots = bb.get();
                return ccid;
            } else {
                bb.position(bb.position() + len - 2);
            }
        }
        return ccid;
    }

    public EnumSet<Feature> getFeatures() { return features; }
}