/*
 * Copyright (c) 2020 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.android.ccid;

import android.util.Log;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.ByteBuffer;

public class CcidT1 extends CcidUSB {

    /**
     * CRC invalid
     */
    public static final int ERR_EDC = -11;


    public static final int RERR_NONE = 0x00;           /* No error indicated in R-block     */
    public static final int RERR_EDC = 0x01;            /* EDC error indicated in  R-block   */
    public static final int RERR_OTHER = 0x02;          /* Other errors in R-block           */
    public static final int RERR_MASK = 0x0F;           /* Mask to decode errors             */

    public static final byte SADMASK = 0x07;            /* Mask for source id                */
    public static final byte DADMASK = 0x70;            /* Mask for destination id           */

    public static final byte NONIBLOCK = (byte) 0x80;   /* Non I Block indicator             */
    public static final byte SBLOCK = 0x40;             /* S Block indicator                 */
    public static final byte NRBIT = 0x10;              /* N(R) indicator                    */
    public static final byte NSBIT = 0x40;              /* N(S) indicator                    */
    public static final byte MOREBIT = (byte) 0x20;     /* More bit indicator                */

    public static final byte SBLOCKFUNCMASK = 0x3F;      /* S-block function mask             */
    public static final byte RESYNCHREQ = 0x00;          /* S-block with RESYNCH request      */
    public static final byte RESYNCHRES  = 0x20;         /* S-block with RESYNCH response     */
    public static final byte IFSREQ = 0x01;              /* S-block with IFS request          */
    public static final byte IFSRES = 0x21;              /* S-block with IFS response         */
    public static final byte ABORTREQ = 0x02;            /* S-block with ABORT request        */
    public static final byte ABORTRES = 0x22;            /* S-block with ABORT response       */
    public static final byte WTXREQ = 0x03;              /* S-block with WTX request          */
    public static final byte WTXRES = 0x23;              /* S-block with WTX response         */

    public static final byte VPPERRRES = 0x24;           /* S-block with VPPERR response      */

    /** Character Wait Time in ETU        */
    int CharWaitTime;
    /** Block Wait Time in ETU            */
    int BlockWaitTime;
    /** Working Block Wait Time in ETU    */
    long WorkBWT;
    /** Receiver sequence number          */
    byte RSequenz;
    /** Transmitter sequence number       */
    byte SSequenz;
    /** NAD unsigned char of received block        */
    byte Nad;
    /** PCB unsigned char of received blocks       */
    byte Pcb;
    /** Length of received data block     */
    short InBuffLength;
    /** Buffer for incoming data          */
    byte[] InBuff = new byte[BUFFMAX];


    public CcidT1(USBDeviceHandler usbHandler) {
        super(usbHandler);
    }

    public static byte CODENAD(byte sad, byte dad) {
        return (byte) (((dad & (byte) 0x0F) << 4) | (sad & (byte) 0x0f));
    }

    public static byte CODESBLOCK(byte x) {
        return (byte) (((x) | NONIBLOCK | SBLOCK));
    }

    public static byte CODERBLOCK(byte nr, byte rc) {
        return (byte) (NONIBLOCK | ((nr) << 4) | (rc));
    }

    public static byte CODEIBLOCK(byte ns, byte mb) {
        return (byte) (((ns) << 6) | ((mb) << 5));
    }

    public static boolean ISIBLOCK(byte x) {
        return ((x & NONIBLOCK) == 0);
    }

    public static boolean ISRBLOCK(byte x) {
        return ((((x) & NONIBLOCK) != 0) && (((x) & SBLOCK) == 0));
    }

    public static boolean ISSBLOCK(byte x) {
        return ((((x) & NONIBLOCK) != 0) && (((x) & SBLOCK) != 0));
    }

    public static byte NR(byte x) {
        return (byte) (((x) & NRBIT) >> 4);
    }

    public static byte NS(byte x) {
        return (byte) (((x) & NSBIT) >> 6);
    }

    public static byte RERR(byte x) {
        return (byte) ((x) & RERR_MASK);
    }

    public static byte MORE(byte x) {
        return (byte) (((x) & MOREBIT) >> 5);
    }

    public static byte SBLOCKFUNC(byte x) {
        return (byte) ((x) & SBLOCKFUNCMASK);
    }

    public static byte SAD(byte x) {
        return (byte) ((x) & SADMASK);
    }

    public static byte DAD(byte x) {
        return (byte) (((x) & DADMASK) >> 4);
    }


    public static String hexify( byte [] raw ) {
        final String HEXES = "0123456789ABCDEF";

        if ( raw == null ) {
            return null;
        }
        final StringBuilder hex = new StringBuilder( 2 * raw.length );
        for ( final byte b : raw ) {
            hex.append(HEXES.charAt((b & 0xF0) >> 4))
                    .append(HEXES.charAt((b & 0x0F)));
        }
        return hex.toString();
    }

    static void ccidT1BlockInfo(byte Nad, byte Pcb, byte[] block)
    {
        Log.d("CCID_T1", String.format("SAD:%02d-DAD:%02d\n", SAD(Nad), DAD(Nad)));

        if (ISIBLOCK(Pcb)) {
            Log.d("CCID_T1", String.format("I(%d,%d):\n", NS(Pcb), MORE(Pcb)));
            Log.d("CCID_T1", hexify(block));
        }

        if (ISRBLOCK(Pcb)) {
            Log.d("CCID_T1", String.format("R(%d)[%d]\n", NR(Pcb), RERR(Pcb)));
        }

        if (ISSBLOCK(Pcb)) {
            switch (SBLOCKFUNC(Pcb)) {
                case RESYNCHREQ :
                    Log.d("CCID_T1", "S(RESYNCH Request)\n");
                    break;
                case RESYNCHRES :
                    Log.d("CCID_T1","S(RESYNCH Response)\n");
                    break;
                case IFSREQ :
                    Log.d("CCID_T1",String.format("S(IFS Request:%d)\n", block != null ? (byte)block[0] : -1));
                    break;
                case IFSRES :
                    Log.d("CCID_T1",String.format("S(IFS Response:%d)\n",block != null ? (byte)block[0] : -1));
                    break;
                case ABORTREQ       :
                    Log.d("CCID_T1","S(ABORT Request)\n");
                    break;
                case ABORTRES       :
                    Log.d("CCID_T1","S(ABORT Response)\n");
                    break;
                case WTXREQ :
                    Log.d("CCID_T1",String.format("S(WTX Request:%d)\n",block != null ? (byte)block[0] : -1));
                    break;
                case WTXRES :
                    Log.d("CCID_T1",String.format("S(WTX Response:%d)\n",block != null ? (byte)block[0] : -1));
                    break;
                default :
                    Log.d("CCID_T1","Unknown S(...) Block\n");
                    break;
            }
        }
    }

    /**
     * Receive a block in T=1 protocol
     *
     */
    int ccidT1ReceiveBlock()
    {
        int rc = 0;
        int i, len;
        byte lrc = 0;
        InBuffLength = (short) -1;
        byte[] block;

        try {
            block = RDR_to_PC_DataBlock(null, null, null);
        } catch (IOException ioe) {
            return -1;
        }

        Log.d("CCID_T1", "Received : ");
        ccidT1BlockInfo(block[0], block[1], ByteBuffer.wrap(block, 2, block.length - 2 ).array());
        Log.d("CCID_T1", "RAW: " + hexify(block));

        if (block.length != 0) {
            /* Calculate checksum */
            for (i = 0; i < (block.length - 1); i++) {
                lrc ^= block[i];
            }

            if (lrc != block[block.length - 1]) {
                return ERR_EDC;
            }
        }

        Nad = block[0];
        Pcb = block[1];
        InBuffLength = (short) (block[2] & 0x00FF);

        if (InBuffLength > 0) {
            System.arraycopy(block, 3, InBuff, 0, InBuffLength);
        }

        return 0;
    }


    void ccidT1SendBlock(byte Nad,
                         byte Pcb,
                         byte[] data, int offset, short length) throws IOException
    {
        int rc, len;
        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        int lrc;

        if (length > 254) {
            throw new IOException("T=1 block too large");
        }

        bos.write(Nad);
        bos.write(Pcb);
        bos.write((byte) (length & 0xFF));
        if (length != 0) {
            bos.write(data, offset, length);
        }

        lrc = 0;

        byte[] scratch = bos.toByteArray();

        for (int i = 0; i < scratch.length; i++) {
            lrc ^= scratch[i];
        }

        bos.write((byte) lrc);

        byte[] block = bos.toByteArray();
        PC_to_RDR_XfrBlock(block, 0, block.length, 0);

        Log.d("CCID_T1", "Sending : ");
        ccidT1BlockInfo(Nad, Pcb, length > 0 ? ByteBuffer.wrap(data, offset, length).array() : null);
        Log.d("CCID_T1", "RAW: " + hexify(block));
    }


    void ccidT1Resynch(byte SrcNode, byte DestNode) throws IOException
    {
        int ret,retry;

        retry = 2;

        while (retry-- > 0) {
            ccidT1SendBlock(CODENAD(SrcNode, DestNode),
                    CODESBLOCK(RESYNCHREQ), null, 0, (short) 0);

            ret = ccidT1ReceiveBlock();

            if (ret < 0 &&
                    ISSBLOCK(Pcb) &&
                    (SBLOCKFUNC(Pcb) == RESYNCHRES)) {
                // ccidT1InitProtocol(ctx);
                return;
            }
        }
    }


    void ccidT1AbortChain(byte SrcNode, byte DestNode) throws IOException
    {
        int ret,retry;

        retry = 2;

        while (retry-- > 0) {
            ccidT1SendBlock(CODENAD(SrcNode, DestNode),
                    CODESBLOCK(ABORTREQ),
                    null, 0, (short) 0);

            ccidT1ReceiveBlock();

        }
    }



    void ccidT1SetIFSD(byte SrcNode, byte DestNode) throws IOException
    {
        byte[] blk = new byte[] { (byte)252 };
        ccidT1SendBlock(CODENAD(SrcNode, DestNode), CODESBLOCK(IFSREQ),
                blk, 0, (short)1);
        ccidT1ReceiveBlock();
    }



    int ccidT1GetBlock(byte SrcNode, byte DestNode) throws IOException
    {
        int retry,ret;

        retry = 2;
        WorkBWT = BlockWaitTime;

        while (true) {
            ret = ccidT1ReceiveBlock();

            if (ret < 0) {
                if (retry == 0) {
                    return -1;
                }

                /* The receiver did not understand our transmission block.             */
                /* Unless the retry counter expires, we send it again                  */

                retry--;
                ccidT1SendBlock(CODENAD(SrcNode, DestNode),
                        CODERBLOCK(RSequenz, (byte) (ret == ERR_EDC ? 1 : 2)),
                        null, 0, (short) 0);

                WorkBWT = BlockWaitTime;
                continue;
            }

            /* Lets see, if we received a S-block from the other side                */

            if (ISSBLOCK(Pcb)) {
                switch(SBLOCKFUNC(Pcb)) {
                    case RESYNCHRES :               /* Request to synchronize again      */
                        // ccidT1InitProtocol(ctx);
                        return 1;

                    case IFSREQ :                   /* Request to change the buffer size */
                        ccidT1SendBlock(CODENAD(SrcNode, DestNode),
                                CODESBLOCK(IFSRES),
                                InBuff, 0, (short) 1);
                        IFSC = InBuff[0];
                        break;

                    case ABORTREQ :                 /* Request to abort the transmission */
                        ccidT1SendBlock(CODENAD(SrcNode, DestNode),
                                CODESBLOCK(ABORTRES),
                                null, 0, (short) 0);

                        ccidT1ReceiveBlock();
                        return -1;

                    case WTXREQ :                   /* Request to extend timeout         */
                        ccidT1SendBlock(CODENAD(SrcNode, DestNode),
                                CODESBLOCK(WTXRES),
                                InBuff, 0, (short) 1);
                        WorkBWT = BlockWaitTime * (int)InBuff[0];
                        break;

                    default :
                        return -1;
                }
            }

            if (ISRBLOCK(Pcb) || ISIBLOCK(Pcb)) {
                break;
            }
        }

        return 0;
    }


    void ccidT1SendData(boolean HostMode,
                        byte SrcNode,
                        byte DestNode,
                        byte[] cmd, short cmdLen) throws IOException
    {
        int ret, retry, length, response;
        byte more;
        short offset = 0;

        while (cmdLen > 0) {
            if (((short) cmd.length - offset) > IFSC) {
                length = IFSC;
            } else {
                length = cmdLen;
            }
            cmdLen -= length;
            more = (byte) (cmdLen > 0 ? 1 : 0);

            // #define RETRY   2                       /* Number of retries                 */
            retry = 2;
            response = 0;

            while (response <= 0) {

                ccidT1SendBlock(CODENAD(SrcNode, DestNode),
                        CODEIBLOCK(SSequenz, more),
                        cmd,
                        offset,
                        (short) (length & 0x00FF));

                if (more == 0 && HostMode) {
                    SSequenz = (byte) (1 - SSequenz);
                    return;
                }

                ret = ccidT1GetBlock(SrcNode, DestNode);

                if (ret > 0) {                     /* Send block again                  */
                    continue;
                }

                /* A block can be acknowledged with an I or R-block                    */

                if (ISRBLOCK(Pcb)) {
                    if (NR(Pcb) == SSequenz) {
                        if (retry-- > 0) {
                            continue;                   /* Receiver is requesting same block */
                        }

                        ccidT1Resynch(SrcNode,DestNode);

                        continue;
                    } else if (more > 0) {

                        /* In chaining mode, the R-block is used to acknowledge the I-block*/

                        response = 1;
                        SSequenz = (byte) (1 - SSequenz);
                    } else {

                        Log.d("CCID_T1", "Error: Unexpected R-Block.\n");
                        ccidT1Resynch(SrcNode,DestNode);
                    }
                }

                if (ISIBLOCK(Pcb)) {    /* Usual response to I-block     */
                    if ((NS(Pcb) != RSequenz) || more > 0) {

                        Log.d("CCID_T1", "Asynchronous I-Block received as response.\n");
                        ccidT1Resynch(SrcNode,DestNode);

                    } else {
                        SSequenz = (byte) (1 - SSequenz);
                        response = 1;
                    }
                }
            }

            offset += length;
        }
    }


    byte[] ccidT1ReceiveData(byte SrcNode,
                             byte DestNode) throws IOException
    {
        int more,Length,ret;

        Length = 0;
        ByteArrayOutputStream bos = new ByteArrayOutputStream();


        do  {

            more = MORE(Pcb);     /* More block following ?            */

            if (InBuffLength > 0) {
                bos.write(InBuff, 0, InBuffLength);
            }

            RSequenz = (byte) (1 - RSequenz);

            if (more > 0) {
                while (true) {
                    ccidT1SendBlock(CODENAD(SrcNode, DestNode),
                            CODERBLOCK(RSequenz, (byte) 0),
                            null, 0, (short) 0);

                    ccidT1GetBlock(SrcNode, DestNode);

                    if (ISRBLOCK(Pcb)) {
                        if (NR(Pcb) != SSequenz) {

                            Log.d("CCID_T1", "Error: Invalid sequenz received in R-Block.\n");
                            ccidT1Resynch(SrcNode, DestNode);
                        }

                        continue;
                    }

                    break;
                }
            }
        } while (more > 0);

        return bos.toByteArray();
    }


    byte[] ccidT1Transport(byte SrcNode, byte DestNode, byte[] cmd) throws IOException {

        ccidT1SendData(false, SrcNode, DestNode, cmd, (short) cmd.length);

        return ccidT1ReceiveData(SrcNode, DestNode);
    }


    public byte[] ccidAPDUProcess (byte[] cmd) throws IOException {
        return  ccidT1Transport( (byte) 0, (byte) 0, cmd);
    }


    public void ccidAPDUInit () throws IOException {
        ccidT1SetIFSD((byte) 0, (byte) 0);
    }
}
