/*
 * Copyright (c) 2020 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.android.ccid;

import java.io.IOException;

public abstract class CcidUSB {

    /**
     * Maximum size of ATR
     */
    protected final static byte MAX_ATR = 34;
    protected final static int BUFFMAX = 261;

    USBDeviceHandler usbh;
    protected byte[] atr;

    /** Clock rate conversion integer      */
    protected byte FI;
    /** Baud rate adjustment integer       */
    protected byte DI;
    /** Character waiting time             */
    protected byte CWI;
    /** Block waiting time                 */
    protected byte BWI;
    /** Extra guard time                   */
    protected byte EXTRA_GUARD_TIME;
    /** Maximum length of INF field        */
    protected short IFSC;
    /** Current baudrate                   */
    protected int Baud;

    //One on one copy from the ccid_usb.c file
    private final int[] FTable = new int[] { 372, 372, 558, 744, 1116, 1488, 1860, -1, -1, 512, 768, 1024, 1536, 2048, -1, -1};
    private final int[] DTable = new int[] { -1, 1, 2, 4, 8, 16, 32, -1, 12, 20, -1, -1, -1, -1, -1, -1};


    public CcidUSB(USBDeviceHandler usbHandler) {
        this.usbh = usbHandler;
    }


    //One on one copy from the ccid_usb.c file
    private boolean match(int x, int y) {

        return ((x >= (y - y / 20)) && (x <= (y + y / 20)));

    }

    int DetermineBaudrate(int F, int D)
    {
        int br;

        br = 14318000 * D / (F * 4);

        if (match(br, 9600)) {
            br = 9600;
        } else if (match(br, 19200)) {
            br = 19200;
        } else if (match(br, 38400)) {
            br = 38400;
        } else if (match(br, 57600)) {
            br = 57600;
        } else if (match(br, 115200)) {
            br = 115200;
        } else {
            br = -1;
        }

        return br;
    }

    void decodeATRValues(byte[] atr) throws IOException
    {

        int atrp;
        byte i;
        byte temp, prot;
        byte help;

        this.FI = 1;
        this.DI = 1;

        this.IFSC = 32;              /* T=1: information field size TA(i)*/
        this.CWI = 13;               /* T=1: Char waiting time indx TB(i)*/
        this.BWI = 4;                /* T=1: Block waiting time inx TB(i)*/

        atrp = 0;

        /* Check initial char in ATR */
        temp = atr[atrp++];

        if (temp != 0x3B && temp != 0x3F) {
            throw new IOException("Invalid card ATR");
        }

        /* Get T0 */
        temp = atr[atrp++];
        // NumOfHB = temp & 0x0F;

        i = 1;

        do {

            prot = (byte) (temp & 0x0F);
            help = (byte) ((temp >> 4) & 0x0F);

            if ((help & 1) == 1) { /* Get TAx                          */
                if (i == 1) { /* TA(1) present ?                */
                    temp = atr[atrp++];
                    this.FI = (byte) (temp >> 4);
                    this.DI = (byte) (temp & 0xF);
                }

                if (i > 2 && (prot != 0x0F)) {
                    temp = atr[atrp++];
                    this.IFSC = (short) (temp & 0x00FF);
                }
            }

            if ((help & 2) == 2) { /* Get TBx                          */
                temp = atr[atrp++];

                if (i > 2  && (prot != 0x0F)) {
                    this.CWI = (byte) (temp & 0x0F);
                    this.BWI = (byte) (temp >> 4);
                }
            }

            if ((help & 4) == 4) { /* Get TCx                          */
                temp = atr[atrp++];

                if (i > 2  && (prot != 0x0F)) { /* TC(1) present ?                  */
                    this.EXTRA_GUARD_TIME = temp;
                }
            }

            if ((help & 8) == 8) { /* Get TDx                          */
                temp = atr[atrp++];
            } else {
                temp = 0;
            }

            i++;

        } while (temp != 0);

        /*
        for (i = 0; i < ctx->NumOfHB; i++) {
            ctx->HCC[i] = ctx->ATR[atrp++];
        }
        */

        this.Baud = DetermineBaudrate(FTable[FI], DTable[DI]);

        PC_to_RDR_SetParameters();
    }

    /**
     * Power on the Smart Card
     * @throws IOException If something goes wrong
     */
    public byte[] PC_to_RDR_IccPowerOn() throws IOException {
        int rc;
        byte[] msg = new byte[10 + MAX_ATR];
        int atrlen, l = 10 + MAX_ATR;

        msg[0] = MessageType.PC_TO_RDR_ICC_POWER_ON.getID();

        rc = usbh.connection.bulkTransfer(usbh.pipeOut, msg, 10, 10000000);

        if (rc < 0) {
            throw new IOException("Error sending data");
        }

        rc = usbh.connection.bulkTransfer(usbh.pipeIn, msg, msg.length, 10000000);

        if (rc < 0) {
            throw new IOException("Error receiving data");
        }

        /* check length, message type, slot and sequence number */
        if (rc < 10 || msg[0] != MessageType.RDR_TO_PC_DATABLOCK.getID() || msg[5] != 0x00 || msg[6] != 0x00) {
            throw new IOException("Error resetting card");
        }

        atrlen = (msg[4] << 24) + (msg[3] << 16) + (msg[2] << 8) + msg[1];

        atr = new byte[atrlen];
        System.arraycopy(msg, 10, atr, 0, atrlen);

        decodeATRValues(atr);
        ccidAPDUInit();

        /*
        rc = PC_to_RDR_SetParameters(ctx);

        if (rc < 0) {
            return rc;
        }

        */

        return atr;
    }


    /**
     * Power off the Smart Card
     * @throws IOException If something goes wrong
     */

    public void PC_to_RDR_IccPowerOff() throws IOException {
        byte[] msg = new byte[10];
        byte[] buf = new byte[10];
        int rc;

        msg[0] = MessageType.PC_TO_RDR_ICC_POWER_OFF.getID();

        rc = usbh.connection.bulkTransfer(usbh.pipeOut, msg, msg.length, 10000000);

        if (rc < 0) {
            throw new IOException("Error sending data");
        }

        rc = usbh.connection.bulkTransfer(usbh.pipeIn, buf, buf.length, 10000000);

        if (rc < 0) {
            throw new IOException("Error receiving data");
        }

        /* check length, message type, slot and sequence number */
        if (rc != 10 || buf[0] != MessageType.RDR_TO_PC_SLOT_STATUS.getID() || buf[5] != 0x00 || buf[6] != 0x00) {
            throw new IOException("Error powering down card");
        }

    }


    public void PC_to_RDR_XfrBlock(byte[] outbuf, int offset, int length, int level) throws IOException {
        int rc;
        byte[] msg = new byte[10 + BUFFMAX];

        if (length > BUFFMAX) {
            throw new IOException("Outgoing buffer too large");
        }

        msg[0] = MessageType.PC_TO_RDR_XFRBLOCK.getID();
        msg[1] = (byte) (length & 0xFF);
        msg[2] = (byte) ((length >> 8) & 0xFF);
        msg[3] = (byte) ((length >> 16) & 0xFF);
        msg[4] = (byte) ((byte) (length >> 24) & 0xFF);
        msg[8] = (byte) (level & 0xFF);
        msg[9] = (byte) ((byte) (level >> 8) & 0xFF);
        if (length > 0) {
            System.arraycopy(outbuf, 0, msg, 10, length);
        }

        rc = usbh.connection.bulkTransfer(usbh.pipeOut, msg, (10 + length), 10000000);

        if (rc < 0) {
            throw new IOException("Error sending data");
        }

    };


    public byte[] RDR_to_PC_DataBlock(byte[] status, byte[] error, byte[] chain) throws IOException {
        byte[] msg = new byte[10 + BUFFMAX];
        int rc;

        while (true) {
            rc = usbh.connection.bulkTransfer(usbh.pipeIn, msg, msg.length, 10000000);

            if (rc < 0) {
                throw new IOException("Error receiving data");
            }

            /* check length, message type, slot and sequence number */
            if (rc < 10 || msg[0] != MessageType.RDR_TO_PC_DATABLOCK.getID() || msg[5] != 0x00 || msg[6] != 0x00) {
                throw new IOException("Error sending data");
            }

            if ((msg[7] & (byte) 0x80) == (byte) 0x80) {			// Card requests waiting time extension
                continue;
            }

            break;
        }

        if (status != null)
                status[0] = msg[7];
        if (error != null)
                error[0] = msg[8];
        if (chain != null)
                chain[0] = msg[9];

        byte[] inbuf = new byte[rc - 10];
        System.arraycopy(msg, 10, inbuf, 0, rc - 10 );
        return inbuf;
    };


    public int PC_to_RDR_GetSlotStatus() throws IOException {
        byte[] msg = new byte[10];
        byte[] buf = new byte[10];
        int slotstatus;
        int rc;

        msg[0] = MessageType.PC_TO_RDR_GET_SLOT_STATUS.getID();

        rc = usbh.connection.bulkTransfer(usbh.pipeOut, msg, msg.length, 10000000);

        if (rc < 0) {
            throw new IOException("Error sending data");
        }

        rc = usbh.connection.bulkTransfer(usbh.pipeIn, buf, buf.length, 10000000);

        if (rc < 0) {
            throw new IOException("Receiving data");
        }

        /* check length, message type, slot and sequence number */
        if (rc != 10 || buf[0] != MessageType.RDR_TO_PC_SLOT_STATUS.getID() || buf[5] != 0x00 || buf[6] != 0x00) {
            throw new IOException("Invalid response in slot status query");
        }

        slotstatus = buf[7];

        /*
        #define ICC_PRESENT_AND_ACTIVE		0x00
        #define ICC_PRESENT_AND_INACTIVE	0x01
        #define NO_ICC_PRESENT				0x02
        #define ICC_STATUS_MASK				0x03
         */

        return (slotstatus & 0x03);
    }


    private void PC_to_RDR_SetParameters() throws IOException {

        byte[] msg = new byte[17];
        byte[] buf = new byte[17];
        int rc;
        int len = 0;

        /* 61 07 00 00 00 00 0F 01 00 00 18 10 02 45 00 FE 00 */
        msg[0] = MessageType.PC_TO_RDR_SET_PARAMETERS.getID();
        msg[1] = 0x07;
        msg[7] = 0x01;  /* T=1 protocol */
        msg[10] = (byte) ((byte) (FI << 4) | (byte) (DI & 0x0F)); /* FI, DI */
        msg[11] = 0x10; /* CRC, direct convention */
        msg[12] = EXTRA_GUARD_TIME; /* Extra guard time */
        msg[13] = (byte) ((byte) (BWI << 4) | (byte) (CWI & 0x0F)); /* BWI, CWI */
        msg[14] = 0x00; /* Stopping clock is not allowed */
        msg[15] = (byte) IFSC; /* Negotiated IFSC = 254 bytes */
        msg[16] = 0x00; /* Default value for NAD */


        rc = usbh.connection.bulkTransfer(usbh.pipeOut, msg, msg.length, 10000000);

        if (rc < 0) {
            throw new IOException("Error sending data");
        }

        rc = usbh.connection.bulkTransfer(usbh.pipeIn, buf, buf.length, 10000000);

        if (rc < 0) {
            throw new IOException("Receiving data");
        }

        /* check length, message type, slot and sequence number */
        if (rc != 17 || buf[0] != MessageType.RDR_TO_PC_PARAMETERS.getID() || buf[5] != 0x00 || buf[6] != 0x00) {
            throw new IOException("Invalid response setting reader parameters");
        }
    }

    public abstract void ccidAPDUInit () throws IOException;
    public abstract byte[] ccidAPDUProcess (byte[] cmd) throws IOException;
}
