/*
 * Copyright (c) 2020 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.android.nfc;

import android.nfc.NfcAdapter;
import android.nfc.Tag;
import android.nfc.TagLostException;
import android.nfc.tech.IsoDep;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;

import opencard.core.terminal.CardID;
import opencard.core.terminal.CardNotPresentException;
import opencard.core.terminal.CardTerminal;
import opencard.core.terminal.CardTerminalException;
import opencard.core.terminal.CommandAPDU;
import opencard.core.terminal.Pollable;
import opencard.core.terminal.ResponseAPDU;
import opencard.core.terminal.SlotChannel;


public class NfcCardTerminal extends CardTerminal implements NfcAdapter.ReaderCallback {

    private final static Logger logger = LoggerFactory.getLogger(NfcCardTerminal.class);

    private NfcAdapter nfcAdapter;
    private IsoDep card = null;
    private int maxAPDU;
    private CardID cardID = null;


    public NfcCardTerminal(String name, String type, String address, NfcAdapter nfcAdapter)
            throws CardTerminalException {
        super(name, type, address);

        logger.debug("Adding NfcCardTerminal");

        this.nfcAdapter = nfcAdapter;
        addSlots(1);
    }



    @Override
    public CardID getCardID(int slotID) throws CardTerminalException {
        logger.debug("getCardID called");

        if (!isCardPresent(slotID)) {
            return null;
        }
        return cardID;
    }



    @Override
    public boolean isCardPresent(int slotID) throws CardTerminalException {
        boolean cardIsPresent = (card != null);
        logger.debug("isCardPresent called: " + Boolean.toString(cardIsPresent));
        return cardIsPresent;
    }



    @Override
    public void open() throws CardTerminalException {
        if (card==null) logger.debug("There is no card yet");
    }



    @Override
    public void close() throws CardTerminalException {
        logger.debug("close()");
    }



    @Override
    protected CardID internalReset(int slot, int ms) throws CardTerminalException {
        if (card == null) {
            logger.debug("No card available");
            throw new CardNotPresentException("No card available", this, 0);
        }

        logger.debug("internalReset(int slot, int ms)");
        return cardID;
    }



    @Override
    protected ResponseAPDU internalSendAPDU(int slot, CommandAPDU capdu, int ms)
            throws CardTerminalException {

        if (card == null) {
            logger.debug("No card available");
            throw new CardNotPresentException("No card available", this, 0);
        }

        if (!card.isConnected()) {
            try {
                card.connect();
            } catch (IOException e) {
                card = null;
                logger.debug("Connect failed with ", e);
                cardRemoved(0);
                throw new CardNotPresentException("Card not present or removed", this, 0);
            }

            if (!card.isConnected()) {
                card = null;
                cardRemoved(0);
                logger.debug("Terminal internalSendAPDU(int slot, CommandAPDU capdu, int ms) - not connected");
                throw new CardNotPresentException("Card not present or removed", this, 0);
            }
        }

        byte[] response = null;

        try {
            response = card.transceive(capdu.getBytes());
        } catch (TagLostException e) {
            card = null;
            logger.debug("Tag lost while transmit", e);
            cardRemoved(0);
            throw new CardTerminalException("Tag lost while transmit " + e.getMessage());
        } catch (IOException e) {
            logger.debug("IOException in transmit ", e);
            throw new CardTerminalException("IOException in transmit  " + e.getMessage());
        }
        return new ResponseAPDU(response);
    }



    @Override
    protected void internalCloseSlotChannel(SlotChannel sc) throws CardTerminalException {
        super.internalCloseSlotChannel(sc);
        card = null;
    }



    private byte[] determineATR(IsoDep card) {
        // get historical bytes for 14443-A
        byte[] historicalBytes = card.getHistoricalBytes();
        if (historicalBytes == null) {
            // get historical bytes for 14443-B
            historicalBytes = card.getHiLayerResponse();
        }
        if (historicalBytes == null) {
            historicalBytes = new byte[0];
        }

        /* copy historical bytes if available */
        byte[] atr = new byte[4 + historicalBytes.length + 1];
        atr[0] = (byte) 0x3b;
        atr[1] = (byte) (0x80 + historicalBytes.length);
        atr[2] = (byte) 0x80;
        atr[3] = (byte) 0x01;
        System.arraycopy(historicalBytes, 0, atr, 4, historicalBytes.length);

        /* calculate TCK */
        byte tck = atr[1];
        for (int idx = 2; idx < atr.length; idx++) {
            tck ^= atr[idx];
        }
        atr[atr.length - 1] = tck;

        return atr;
    }



    public void setTag(Tag t) {
        if (card != null) {
            card = null;
            cardRemoved(0);
        }

        IsoDep sc = IsoDep.get(t);
        try {
            sc.connect();

            cardID = new CardID( this, 0, determineATR(sc));

            if (sc.isExtendedLengthApduSupported()) {
                logger.debug("Extended APDUs supported: good");
            } else {
                logger.debug("Extended APDUs NOT supported: Your milage may vary");
            }

            maxAPDU = sc.getMaxTransceiveLength();
            sc.setTimeout(120000);
            card = sc;
            cardInserted(0);
        } catch (IOException e) {
            logger.debug("setTag failed: " + e.getMessage());
        }
    }



    @Override
    public void onTagDiscovered(Tag tag) {
        logger.debug("Tag discovered");

        String[] techList = tag.getTechList();
        for (String aTechList : techList) {
            if (aTechList.equals("android.nfc.tech.NfcA")) {
                logger.debug("Discovered ISO/IEC 14443-A tag");
            } else if (aTechList.equals("android.nfc.tech.NfcB")) {
                logger.debug("Discovered ISO/IEC 14443-B tag");
            }
        }
        setTag(tag);
    }
} // class
