/*
 * Copyright (c) 2020 CardContact Systems GmbH, Minden, Germany.
 *
 * Redistribution and use in source (source code) and binary (object code)
 * forms, with or without modification, are permitted provided that the
 * following conditions are met:
 * 1. Redistributed source code must retain the above copyright notice, this
 * list of conditions and the disclaimer below.
 * 2. Redistributed object code must reproduce the above copyright notice,
 * this list of conditions and the disclaimer below in the documentation
 * and/or other materials provided with the distribution.
 * 3. The name of CardContact may not be used to endorse or promote products derived
 * from this software or in any other form without specific prior written
 * permission from CardContact.
 * 4. Redistribution of any modified code must be labeled "Code derived from
 * the original OpenCard Framework".
 *
 * THIS SOFTWARE IS PROVIDED BY CardContact "AS IS" FREE OF CHARGE. CardContact SHALL NOT BE
 * LIABLE FOR INFRINGEMENTS OF THIRD PARTIES RIGHTS BASED ON THIS SOFTWARE.  ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  CardContact DOES NOT WARRANT THAT THE FUNCTIONS CONTAINED IN THIS
 * SOFTWARE WILL MEET THE USER'S REQUIREMENTS OR THAT THE OPERATION OF IT WILL
 * BE UNINTERRUPTED OR ERROR-FREE.  IN NO EVENT, UNLESS REQUIRED BY APPLICABLE
 * LAW, SHALL CardContact BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  ALSO, CardContact IS UNDER NO OBLIGATION
 * TO MAINTAIN, CORRECT, UPDATE, CHANGE, MODIFY, OR OTHERWISE SUPPORT THIS
 * SOFTWARE.
 */

package de.cardcontact.opencard.android.swissbit;

import android.content.Context;

import com.secureflashcard.sfclibrary.SfcTerminal;
import com.secureflashcard.sfclibrary.exceptions.TerminalException;
import com.secureflashcard.sfclibrary.exceptions.TerminalNotConnectedException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Properties;

import opencard.core.terminal.CardID;
import opencard.core.terminal.CardTerminal;
import opencard.core.terminal.CardTerminalException;
import opencard.core.terminal.CommandAPDU;
import opencard.core.terminal.ResponseAPDU;

/**
 * Class implementing a card terminal for Swissbit micro SD card
 *
 * @author Mathis Block
 */
public class SBMicroSDCardTerminal extends CardTerminal {

    private final static Logger logger = LoggerFactory.getLogger(SBMicroSDCardTerminal.class);

    private SfcTerminal sfct;

    /**
     * Create a new SBMicroSDCardTerminal
     * @param name Name of the Terminal
     * @param type Type of the Terminal
     * @param address Address of the Terminal
     * @param context Android Application Context
     * @throws CardTerminalException
     * @throws TerminalException
     */
    public SBMicroSDCardTerminal(String name, String type, String address, Context context) throws CardTerminalException, TerminalException {
        super(name,type,address);

        logger.debug("Creating terminal");

        try {
            sfct = SfcTerminal.getInstance(context);
            sfct.connect();
        } catch(TerminalException te) {
            logger.debug("SfcTerminal connect() failed", te);
            throw new CardTerminalException(te.getMessage());
        }

        addSlots(1);
    }



    @Override
    public CardID getCardID(int slotID) throws CardTerminalException {
        byte[] atr = null;

        try {
            atr = sfct.getAtr();
        } catch (TerminalNotConnectedException tnce) {
            logger.error("SfcTerminal.getAtr() failed", tnce);
            return null;
        }

        return new CardID(this, 0, atr);
    }



    @Deprecated
    @Override
    public boolean isCardPresent(int slotID) throws CardTerminalException {
        logger.trace("isCardPresent()");
        return true;
    }



    @Override
    public void open() throws CardTerminalException {
        logger.trace("open()");
    }



    @Override
    public void close() throws CardTerminalException {
        logger.trace("close()");
        sfct.disconnect();
    }



    @Override
    protected CardID internalReset(int slot, int ms) throws CardTerminalException {
        logger.trace("internalReset()");

        byte[] atr = null;
        try {
            atr = sfct.getAtr();
        } catch(TerminalException te) {
            logger.debug("SfcTerminal.getAtr() failed");
            throw new CardTerminalException(te.getMessage());
        }
        return new CardID(this, 0, atr);
    }



    @Override
    protected ResponseAPDU internalSendAPDU(int slot, CommandAPDU capdu, int ms)
            throws CardTerminalException {

        byte[] cmd = capdu.getBytes();
        byte[] rsp = null;

        logger.trace("internalSendAPDU()");

        try {
            rsp = sfct.transmit(cmd);

        } catch(TerminalException te) {
            logger.debug("SfcTerminal.transmit() failed");
            throw new CardTerminalException(te.getMessage());
        }

        return new ResponseAPDU(rsp);
    }



    @Override
    protected Properties internalFeatures(Properties features) {
        features = super.internalFeatures(features);
        features.put("maxCAPDUSize", "450"); // should be 478 if OCF bug is fixed
        features.put("maxRAPDUSize", "506");

        return features;
    }
}
