/*
 *  ---------
 * |.##> <##.|  Open Smart Card Development Platform (www.openscdp.org)
 * |#       #|
 * |#       #|  Copyright (c) 1999-2016 CardContact Systems GmbH
 * |'##> <##'|  32429 Minden, Germany (www.cardcontact.de)
 *  ---------
 *
 *  This file is part of OpenSCDP.
 *
 *  OpenSCDP is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  OpenSCDP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with OpenSCDP; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.cardcontact.jceprovdemo;

import java.io.IOException;
import java.math.BigInteger;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.Security;
import java.security.SignatureException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.security.spec.ECGenParameterSpec;
import java.util.Calendar;
import java.util.Date;

import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginException;

import org.bouncycastle.asn1.x500.X500Name;
import org.bouncycastle.asn1.x509.Extension;
import org.bouncycastle.asn1.x509.SubjectKeyIdentifier;
import org.bouncycastle.cert.X509CertificateHolder;
import org.bouncycastle.cert.jcajce.JcaX509CertificateConverter;
import org.bouncycastle.cert.jcajce.JcaX509ExtensionUtils;
import org.bouncycastle.cert.jcajce.JcaX509v3CertificateBuilder;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.operator.ContentSigner;
import org.bouncycastle.operator.OperatorCreationException;
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder;

import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMECPrivateKeySpec;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMRSAPrivateKeySpec;
import de.cardcontact.smartcardhsmprovider.SmartCardHSMParameterSpec;
import de.cardcontact.smartcardhsmprovider.SmartCardHSMProvider;

/*
 * This demo creates an self signed certificate
 * and stores it into the SmartCardHSM
 */

public class X509Demo {

	public static X509Certificate createSelfSignedCert(KeyPair kp, String algorithm) {

		// Signers name
		String issuer = "C=DE, O=CardContact, CN=Test Client Certificate";

		// On a self signed certificate issuer and subject are the same
		String subject = issuer;

		// Create the certificate

		long ts = System.currentTimeMillis();
		BigInteger serial = BigInteger.valueOf(ts);
		Date from = new Date(ts);
		Calendar cal = Calendar.getInstance();
		cal.setTime(from);
		cal.add(Calendar.YEAR, 5);
		Date to = cal.getTime();

		JcaX509v3CertificateBuilder builder = new JcaX509v3CertificateBuilder(new X500Name(issuer), serial, from, to, new X500Name(subject), kp.getPublic());

		try	{
			JcaX509ExtensionUtils eu = new JcaX509ExtensionUtils();
			SubjectKeyIdentifier skid = eu.createSubjectKeyIdentifier(kp.getPublic());
			builder.addExtension(Extension.subjectKeyIdentifier, true, skid);
		}
		catch(Exception e) {
			e.printStackTrace();
			return null;
		}

		ContentSigner signer = null;
		try {
			signer = new JcaContentSignerBuilder(algorithm).build(kp.getPrivate());
		} catch (OperatorCreationException e) {
			e.printStackTrace();
			return null;
		}

		X509CertificateHolder certholder = builder.build(signer);

		JcaX509CertificateConverter conv = new JcaX509CertificateConverter();

		X509Certificate cert;
		try	{
			cert = conv.getCertificate(certholder);

			cert.checkValidity(new Date());

			cert.verify(kp.getPublic());
		}
		catch(Exception e) {
			e.printStackTrace();
			return null;
		}

		return cert;
	}



	public static KeyPair generateRSAKeyPair() throws NoSuchAlgorithmException, NoSuchProviderException, InvalidAlgorithmParameterException {

		int keySize = 2048;

		SmartCardHSMRSAPrivateKeySpec rsaSpec = new SmartCardHSMRSAPrivateKeySpec(keySize);

		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec("MyRSA", rsaSpec, null);
		KeyPairGenerator kpg;
		KeyPair kp = null;

		kpg = KeyPairGenerator.getInstance("RSA//PKCS1-v1-5-SHA-1", "SmartCardHSM");
		kpg.initialize(params, null);
		kp = kpg.generateKeyPair();

		return kp;
	}



	public static KeyPair generateECKeyPair() throws NoSuchAlgorithmException, NoSuchProviderException, InvalidAlgorithmParameterException {

		SmartCardHSMECPrivateKeySpec ecSpec = new SmartCardHSMECPrivateKeySpec(new ECGenParameterSpec("secp256r1"));

		KeyPairGenerator kpg;
		KeyPair kp = null;
		SmartCardHSMParameterSpec hsmParam = new SmartCardHSMParameterSpec("MyECDSA", ecSpec, null);

		kpg = KeyPairGenerator.getInstance("EC//ECDSA-SHA-1", "SmartCardHSM");
		kpg.initialize(hsmParam, null);
		kp = kpg.generateKeyPair();

		return kp;
	}



	public static void main(String[] args) throws KeyStoreException, NoSuchProviderException, NoSuchAlgorithmException, CertificateException, IOException, InvalidKeyException, IllegalStateException, SignatureException, InvalidAlgorithmParameterException {
		MyPasswordCallbackHandler pch = new MyPasswordCallbackHandler();
		SmartCardHSMProvider hsm = new SmartCardHSMProvider();

		// Make sure to add the BC provider first
		Security.insertProviderAt(new BouncyCastleProvider(), 1);

		try {
			// Login performs a password verification
			hsm.login(null, (CallbackHandler) pch);
		} catch (LoginException e) {
			e.printStackTrace();
			System.exit(-1);
		}

		Security.insertProviderAt(hsm, 1);

		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);

		// RSA
		X509Certificate cert;
		KeyPair kp = generateRSAKeyPair();
		cert = createSelfSignedCert(kp, "SHA256withRSA");
		System.out.println(cert);
		ks.setCertificateEntry("MyRSA", cert);

		// ECC
		kp = generateECKeyPair();
		cert = createSelfSignedCert(kp, "SHA256withECDSA");
		ks.setCertificateEntry("MyECDSA", cert);
	}
}
