/*
 *  ---------
 * |.##> <##.|  Open Smart Card Development Platform (www.openscdp.org)
 * |#       #|
 * |#       #|  Copyright (c) 1999-2016 CardContact Systems GmbH
 * |'##> <##'|  32429 Minden, Germany (www.cardcontact.de)
 *  ---------
 *
 *  This file is part of OpenSCDP.
 *
 *  OpenSCDP is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  OpenSCDP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with OpenSCDP; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.cardcontact.jceprovtest;

import java.io.IOException;
import java.math.BigInteger;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.Security;
import java.security.Signature;
import java.security.SignatureException;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.spec.ECField;
import java.security.spec.ECFieldFp;
import java.security.spec.ECParameterSpec;
import java.security.spec.ECPoint;
import java.security.spec.EllipticCurve;
import java.security.spec.InvalidKeySpecException;
import java.util.Enumeration;
import java.util.stream.Stream;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.KeyAgreement;
import javax.crypto.KeyGenerator;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.SecretKey;
import javax.crypto.ShortBufferException;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.LoginException;

import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.junit.jupiter.api.*;
import org.junit.jupiter.params.*;
import org.junit.jupiter.params.provider.*;

import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMECPrivateKeySpec;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMKey;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMPrivateKey;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMPrivateKeySpec;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMRSAKey;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMRSAPrivateKeySpec;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMSecretKey;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMSecretKeySpec;
import de.cardcontact.smartcardhsmprovider.SmartCardHSMParameterSpec;
import de.cardcontact.smartcardhsmprovider.SmartCardHSMProvider;

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
public class SmartCardHSMProviderTest {

	private SmartCardHSMPrivateKeySpec rsaSpec1024 = new SmartCardHSMRSAPrivateKeySpec(
			"DECA00000", "DEtscons.de00000", 65537, 1024);
	private SmartCardHSMPrivateKeySpec rsaSpec1536 = new SmartCardHSMRSAPrivateKeySpec(
			"DECA00000", "DEtscons.de00001", 65537, 1536);
	private SmartCardHSMPrivateKeySpec rsaSpec2048 = new SmartCardHSMRSAPrivateKeySpec(
			"DECA00000", "DEtscons.de00002", 65537, 2048);

	private ECParameterSpec ecparam;

	private SmartCardHSMPrivateKeySpec ecSpec;

	{
		BigInteger prime = new BigInteger(
				"A9FB57DBA1EEA9BC3E660A909D838D726E3BF623D52620282013481D1F6E5377",	16);
		ECField field = new ECFieldFp(prime);
		BigInteger a, b;
		a = new BigInteger(
				"7D5A0975FC2C3057EEF67530417AFFE7FB8055C126DC5C6CE94A4B44F330B5D9",	16);
		b = new BigInteger(
				"26DC5C6CE94A4B44F330B5D9BBD77CBF958416295CF7E1CE6BCCDC18FF8C07B6",	16);
		EllipticCurve curve = new EllipticCurve(field, a, b);

		BigInteger x = new BigInteger(
				"8BD2AEB9CB7E57CB2C4B482FFC81B7AFB9DE27E1E3BD23C23A4453BD9ACE3262",	16);
		BigInteger y = new BigInteger(
				"547EF835C3DAC4FD97F8461A14611DC9C27745132DED8E545C1D54C72F046997",	16);
		ECPoint g = new ECPoint(x, y);
		BigInteger n = new BigInteger(
				"A9FB57DBA1EEA9BC3E660A909D838D718C397AA3B561A6F7901E0E82974856A7",	16);
		int h = 1;
		ecparam = new ECParameterSpec(curve, g, n, h);

		// String algorithm = "id-TA-ECDSA-SHA-256";
		byte[] algorithm = { 0x04, 0x0, 0x7F, 0x0, 0x07, 0x02, 0x02, 0x02,	0x02, 0x03 };

		this.ecSpec = new SmartCardHSMECPrivateKeySpec("DECA00000", "DEtscons.de00000", ecparam);
		this.ecSpec.setAlgorithm(algorithm);

	}

	@BeforeAll
	public void init() {

		Security.insertProviderAt(new BouncyCastleProvider(), 1);

		MyPasswordCallbackHandler pch = new MyPasswordCallbackHandler();

		SmartCardHSMProvider provider = new SmartCardHSMProvider();
		// provider.setSecureMessaging(false);
		// provider.useTerminalPinPad(true);

		try {
			provider.login(null, pch);
		} catch (LoginException e) {
			e.printStackTrace();
		}

		Security.addProvider(provider);

	}



	@Test
	public void secureRandomTest() throws NoSuchAlgorithmException,
	NoSuchProviderException {
		SecureRandom rng = SecureRandom.getInstance("NativePRNG",
				"SmartCardHSM");

		byte[] rndBytes = new byte[10];
		rng.nextBytes(rndBytes);

		boolean changes = false;
		for (int i = 0; i < rndBytes.length; i++) {
			changes = changes || rndBytes[i] != 0;
		}
		assert (changes);
	}



	@Test
	public void cipherApiTest() throws NoSuchAlgorithmException, NoSuchProviderException, InvalidAlgorithmParameterException, KeyStoreException, UnrecoverableKeyException, CertificateException, IOException, NoSuchPaddingException, InvalidKeyException, IllegalBlockSizeException, BadPaddingException, ShortBufferException {
		// Generate Key Pair
		KeyPairGenerator kpg = KeyPairGenerator.getInstance("RSA", "SmartCardHSM");
		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec(
				"MasterKey", rsaSpec1024);
		kpg.initialize(params);
		KeyPair kp = kpg.generateKeyPair();

		// Get key from KeyStore
		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);
		SmartCardHSMKey prk = (SmartCardHSMKey) ks.getKey("MasterKey", null);
		PublicKey pub = ks.getCertificate("MasterKey").getPublicKey();

		byte[] plain = "Hello World".getBytes();

		Cipher bcCipher = Cipher.getInstance("RSA/NONE/OAEPWithSHA1AndMGF1Padding", "BC");
		bcCipher.init(Cipher.ENCRYPT_MODE, pub);
		bcCipher.update(plain);
		byte[] oaepCipherText = bcCipher.doFinal();
		//byte[] oaepCipherText = bcCipher.doFinal(plain);
		byte[] tmp = new byte[128];

		try {
			Cipher hsmOAEPCipher = Cipher.getInstance("RSA/ECB/OAEPWithSHA-1AndMGF1Padding", "SmartCardHSM");

			hsmOAEPCipher.init(Cipher.DECRYPT_MODE, prk);
			byte[] result = hsmOAEPCipher.update(oaepCipherText);
			result = hsmOAEPCipher.doFinal();
			assert (java.util.Arrays.equals(plain, result));

			hsmOAEPCipher.init(Cipher.DECRYPT_MODE, prk);
			byte[] part1 = new byte[64];
			byte[] part2 = new byte[64];
			System.arraycopy(oaepCipherText, 0, part1, 0, 64);
			System.arraycopy(oaepCipherText, 64, part2, 0, 64);
			result = hsmOAEPCipher.update(part1);
			result = hsmOAEPCipher.doFinal(part2);
			assert (java.util.Arrays.equals(plain, result));

			hsmOAEPCipher.init(Cipher.DECRYPT_MODE, prk);
			result = hsmOAEPCipher.update(oaepCipherText, 0, oaepCipherText.length);
			result = hsmOAEPCipher.doFinal();
			assert (java.util.Arrays.equals(plain, result));

			hsmOAEPCipher.init(Cipher.DECRYPT_MODE, prk);
			int res = hsmOAEPCipher.update(oaepCipherText, 0, oaepCipherText.length, result, 0);
			result = hsmOAEPCipher.doFinal();
			assert (java.util.Arrays.equals(plain, result));
		} finally {
			// Delete key
			ks.deleteEntry("MasterKey");
		}
	}

	Stream<Arguments> rsaKeyGen() {
		return Stream.of(
					Arguments.of(rsaSpec1024),
					Arguments.of(rsaSpec1536),
					Arguments.of(rsaSpec2048)
				);
	}

	@ParameterizedTest
	@MethodSource("rsaKeyGen")
	public void rsaKeyGenTest(SmartCardHSMPrivateKeySpec spec) throws NoSuchAlgorithmException,
	NoSuchProviderException, InvalidAlgorithmParameterException, KeyStoreException, CertificateException, IOException {
		KeyPairGenerator kpg;
		KeyPair kp;
		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec(
				"MasterKey", spec);

		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);

		kpg = KeyPairGenerator.getInstance("RSA", "SmartCardHSM");
		kpg.initialize(params, null);
		kp = kpg.generateKeyPair();
		assert (kp != null);
		assert(kp.getPublic().getAlgorithm().equals(SmartCardHSMKey.RSA));
		assert(kp.getPrivate().getAlgorithm().equals(SmartCardHSMKey.RSA));
		ks.deleteEntry("MasterKey");

		kpg = KeyPairGenerator.getInstance("RSA//PKCS1-v1-5-SHA-1",	"SmartCardHSM");
		kpg.initialize(params, null);
		kp = null;
		kp = kpg.generateKeyPair();
		assert (kp != null);
		assert(kp.getPublic().getAlgorithm().equals(SmartCardHSMKey.RSA));
		assert(kp.getPrivate().getAlgorithm().equals(SmartCardHSMKey.RSA));
		ks.deleteEntry("MasterKey");

		kpg = KeyPairGenerator.getInstance("RSA//PKCS1-v1-5-SHA-256", "SmartCardHSM");
		kpg.initialize(params, null);
		kp = null;
		kp = kpg.generateKeyPair();
		assert (kp != null);
		assert(kp.getPublic().getAlgorithm().equals(SmartCardHSMKey.RSA));
		assert(kp.getPrivate().getAlgorithm().equals(SmartCardHSMKey.RSA));
		ks.deleteEntry("MasterKey");

		kpg = KeyPairGenerator.getInstance("RSA//PKCS1-PSS-SHA-1", "SmartCardHSM");
		kpg.initialize(params, null);
		kp = null;
		kp = kpg.generateKeyPair();
		assert (kp != null);
		assert(kp.getPublic().getAlgorithm().equals(SmartCardHSMKey.RSA));
		assert(kp.getPrivate().getAlgorithm().equals(SmartCardHSMKey.RSA));
		ks.deleteEntry("MasterKey");

		kpg = KeyPairGenerator.getInstance("RSA//PKCS1-PSS-SHA-256", "SmartCardHSM");
		kpg.initialize(params, null);
		kp = null;
		kp = kpg.generateKeyPair();
		assert (kp != null);
		assert(kp.getPublic().getAlgorithm().equals(SmartCardHSMKey.RSA));
		assert(kp.getPrivate().getAlgorithm().equals(SmartCardHSMKey.RSA));
		ks.deleteEntry("MasterKey");
	}



	@Test
	public void ecKeyGenTest() throws NoSuchAlgorithmException,
	NoSuchProviderException, InvalidAlgorithmParameterException, KeyStoreException, CertificateException, IOException {
		KeyPairGenerator kpg;
		KeyPair kp;
		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec(
				"myECDSA", ecSpec);

		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);

		kpg = KeyPairGenerator.getInstance("EC", "SmartCardHSM");
		kpg.initialize(params, null);
		kp = kpg.generateKeyPair();
		assert(kp != null);
		assert(kp.getPublic().getAlgorithm().equals(SmartCardHSMKey.EC));
		assert(kp.getPrivate().getAlgorithm().equals(SmartCardHSMKey.EC));
		ks.deleteEntry("myECDSA");

		kpg = KeyPairGenerator.getInstance("EC//ECDSA-SHA-1", "SmartCardHSM");
		kpg.initialize(params, null);
		kp = null;
		kp = kpg.generateKeyPair();
		assert(kp != null);
		assert(kp.getPublic().getAlgorithm().equals(SmartCardHSMKey.EC));
		assert(kp.getPrivate().getAlgorithm().equals(SmartCardHSMKey.EC));
		ks.deleteEntry("myECDSA");

		kpg = KeyPairGenerator.getInstance("EC//ECDSA-SHA-224", "SmartCardHSM");
		kpg.initialize(params, null);
		kp = null;
		kp = kpg.generateKeyPair();
		assert (kp != null);
		assert(kp.getPublic().getAlgorithm().equals(SmartCardHSMKey.EC));
		assert(kp.getPrivate().getAlgorithm().equals(SmartCardHSMKey.EC));
		ks.deleteEntry("myECDSA");

		kpg = KeyPairGenerator.getInstance("EC//ECDSA-SHA-256", "SmartCardHSM");
		kpg.initialize(params, null);
		kp = null;
		kp = kpg.generateKeyPair();
		assert (kp != null);
		assert(kp.getPublic().getAlgorithm().equals(SmartCardHSMKey.EC));
		assert(kp.getPrivate().getAlgorithm().equals(SmartCardHSMKey.EC));
		ks.deleteEntry("myECDSA");
	}

	Stream<Arguments> rsaSign() {
		return Stream.of(
					Arguments.of(rsaSpec1024, "SHA1withRSA"),
					Arguments.of(rsaSpec1536, "SHA1withRSA"),
					Arguments.of(rsaSpec2048, "SHA1withRSA"),
					Arguments.of(rsaSpec2048, "SHA256withRSA"),
					Arguments.of(rsaSpec1024, "SHA1withRSA/PSS"),
					Arguments.of(rsaSpec1536, "SHA1withRSA/PSS"),
					Arguments.of(rsaSpec2048, "SHA1withRSA/PSS"),
					Arguments.of(rsaSpec2048, "SHA256withRSA/PSS")
				);
	}


	@ParameterizedTest
	@MethodSource("rsaSign")
	public void signatureRSA(SmartCardHSMRSAPrivateKeySpec spec, String algorithm) throws NoSuchAlgorithmException,
	NoSuchProviderException, SignatureException, InvalidKeyException,
	InvalidAlgorithmParameterException, InvalidKeySpecException, KeyStoreException, CertificateException, IOException {

		// Generate random message
		SecureRandom rng = SecureRandom.getInstance("NativePRNG", "SmartCardHSM");
		byte[] b = new byte[100];
		rng.nextBytes(b);

		// Generate key pair
		KeyPairGenerator kpg = KeyPairGenerator.getInstance("RSA", "SmartCardHSM");
		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec("MasterKey", spec);
		kpg.initialize(params, null);
		KeyPair kp = kpg.generateKeyPair();

		Signature signature = Signature.getInstance(algorithm, "SmartCardHSM");
		SmartCardHSMRSAKey key = (SmartCardHSMRSAKey) kp.getPrivate();
		signature.initSign(key);
		signature.update(b, 0, b.length);
		byte[] signedData = signature.sign();

		// Verify message
		Signature verifier = Signature.getInstance(algorithm, "BC");
		verifier.initVerify(kp.getPublic());
		verifier.update(b);
		boolean sigValid = verifier.verify(signedData);
		assert (sigValid);

		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);
		ks.deleteEntry("MasterKey");
	}



	@Test
	public void signatureSHA1withECDSA() throws NoSuchAlgorithmException,
	NoSuchProviderException, InvalidAlgorithmParameterException,
	KeyStoreException, CertificateException, IOException,
	UnrecoverableKeyException, InvalidKeyException, SignatureException {
		// Generate random message
		SecureRandom rng = SecureRandom.getInstance("NativePRNG",
				"SmartCardHSM");
		byte[] message = new byte[100];
		rng.nextBytes(message);

		// Generate key pair
		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec("myECDSA", ecSpec);
		KeyPairGenerator kpg = KeyPairGenerator.getInstance("EC//ECDSA-SHA-1", "SmartCardHSM");
		kpg.initialize(params, null);
		KeyPair kp = kpg.generateKeyPair();

		// Get key from KeyStore
		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);
		Enumeration<String> aliases = ks.aliases();
		SmartCardHSMPrivateKey key = (SmartCardHSMPrivateKey) ks.getKey("myECDSA", null);
		assert(key.getAlgorithm().equals(SmartCardHSMKey.EC));


		// Sign message
		Signature signature = Signature.getInstance("SHA1withECDSA", "SmartCardHSM");
		signature.initSign(key);
		signature.update(message, 0, message.length);
		byte[] signedData = signature.sign();

		// Verify message
		Certificate pukCert = ks.getCertificate("myECDSA");
		PublicKey puk = pukCert.getPublicKey();
		Signature verifier = Signature.getInstance("SHA1withECDSA", "BC");
		verifier.initVerify(puk);
		verifier.update(message);
		boolean sigValid = verifier.verify(signedData);

		assert (sigValid);

		ks.deleteEntry("myECDSA");
	}



	@Test
	public void signatureNONEwithECDSA() throws NoSuchAlgorithmException,
	NoSuchProviderException, InvalidAlgorithmParameterException,
	InvalidKeyException, SignatureException, KeyStoreException,
	CertificateException, IOException, UnrecoverableKeyException {
		// Generate random message
		SecureRandom rng = SecureRandom.getInstance("NativePRNG", "SmartCardHSM");
		byte[] message = new byte[100];
		rng.nextBytes(message);

		// Create Message Digest
		MessageDigest sha256 = MessageDigest.getInstance("SHA256", "BC");
		sha256.update(message);
		byte[] digest = sha256.digest();

		// Generate key pair
		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec(
				"myECDSA", ecSpec);
		KeyPairGenerator kpg = KeyPairGenerator.getInstance("EC//ECDSA-SHA-1", "SmartCardHSM");
		kpg.initialize(params, null);
		KeyPair kp = kpg.generateKeyPair();

		// Get key from KeyStore
		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);
		Enumeration<String> aliases = ks.aliases();
		SmartCardHSMPrivateKey key = (SmartCardHSMPrivateKey) ks.getKey("myECDSA", null);
		assert(key.getAlgorithm().equals(SmartCardHSMKey.EC));

		Certificate pukCert = ks.getCertificate("myECDSA");
		PublicKey puk = pukCert.getPublicKey();

		// Sign message
		Signature signature = Signature.getInstance("NONEwithECDSA", "SmartCardHSM");
		signature.initSign(key);
		signature.update(digest, 0, digest.length);
		byte[] signedData = signature.sign();

		// Verify message
		Signature verifier = Signature.getInstance("NONEwithECDSA", "BC");
		verifier.initVerify(puk);
		verifier.update(digest);
		boolean sigValid = verifier.verify(signedData);
		assert (sigValid);

		ks.deleteEntry("myECDSA");
	}



	@Test
	public void deleteEntryTest() throws KeyStoreException,
	NoSuchProviderException, NoSuchAlgorithmException,
	CertificateException, IOException, UnrecoverableKeyException,
	InvalidKeySpecException, InvalidKeyException, SignatureException,
	InvalidAlgorithmParameterException {
		boolean deleted = false;
		boolean notFound = false;

		// Generate EC Key Pair
		KeyPairGenerator kpg = KeyPairGenerator.getInstance("EC", "SmartCardHSM");
		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec("DeleteMe", ecSpec);
		kpg.initialize(params, null);
		KeyPair kp = kpg.generateKeyPair();

		// Get key from KeyStore
		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore",
				"SmartCardHSM");
		ks.load(null, null);
		Enumeration<String> aliases = ks.aliases();
		SmartCardHSMKey key = (SmartCardHSMKey) ks.getKey("DeleteMe", null);

		ks.deleteEntry("DeleteMe");

		try {
			ks.deleteEntry("DeleteMe2");
		} catch (KeyStoreException e) {
			notFound = true;
		}
		assert (notFound == true);

		try {
			SmartCardHSMKey key2 = (SmartCardHSMKey) ks.getKey("DeleteMe", null);
		} catch (UnrecoverableKeyException e) {
			deleted = true;
		}
		assert (deleted == true);
	}



	@Test
	public void testAllEngines() throws KeyStoreException,
	NoSuchProviderException, NoSuchAlgorithmException,
	CertificateException, IOException, UnrecoverableKeyException,
	InvalidKeySpecException, InvalidKeyException, SignatureException,
	InvalidAlgorithmParameterException {

		// Generate random message
		SecureRandom rng = SecureRandom.getInstance("NativePRNG", "SmartCardHSM");
		byte[] message = new byte[100];
		rng.nextBytes(message);

		// Generate EC Key Pair
		KeyPairGenerator kpg = KeyPairGenerator.getInstance("EC", "SmartCardHSM");
		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec("NewEC", ecSpec);
		kpg.initialize(params, null);
		KeyPair kp = kpg.generateKeyPair();

		// Get key from KeyStore
		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);
		Enumeration<String> aliases = ks.aliases();
		SmartCardHSMPrivateKey key = (SmartCardHSMPrivateKey) ks.getKey("NewEC", null);
		assert(key.getAlgorithm().equals(SmartCardHSMKey.EC));

		Certificate pukCert = ks.getCertificate("NewEC");
		PublicKey puk = pukCert.getPublicKey();

		// Sign message
		Signature signature = Signature.getInstance("SHA1withECDSA", "SmartCardHSM");
		signature.initSign(key);
		signature.update(message, 0, message.length);
		byte[] signedData = signature.sign();

		// Verify message
		Signature verifier = Signature.getInstance("SHA1withECDSA", "BC");
		verifier.initVerify(puk);
		verifier.update(message);
		boolean sigValid = verifier.verify(signedData);
		assert(sigValid);

		// Delete key
		ks.deleteEntry("NewEC");
		aliases = ks.aliases();
	}

	@ParameterizedTest
	@MethodSource("rsaKeyGen")
	public void cipher(SmartCardHSMRSAPrivateKeySpec spec) throws KeyStoreException, NoSuchProviderException,
	NoSuchAlgorithmException, CertificateException, IOException,
	UnrecoverableKeyException, InvalidKeySpecException,
	InvalidKeyException, SignatureException,
	InvalidAlgorithmParameterException, NoSuchPaddingException,
	IllegalBlockSizeException, BadPaddingException {

		// Generate Key Pair
		KeyPairGenerator kpg = KeyPairGenerator.getInstance("RSA", "SmartCardHSM");
		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec(
				"MasterKey", spec);
		kpg.initialize(params, null);
		KeyPair kp = kpg.generateKeyPair();

		// Get key from KeyStore
		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);
		SmartCardHSMKey prk = (SmartCardHSMKey) ks.getKey("MasterKey", null);
		assert(prk.getAlgorithm().equals(SmartCardHSMKey.RSA));
		PublicKey pub = ks.getCertificate("MasterKey").getPublicKey();
		assert(pub.getAlgorithm().equals(SmartCardHSMKey.RSA));

		try {
			// Encrypt
			byte[] plain = "Hello World".getBytes();
			byte[] cryptogram = new byte[prk.getKeySize() / 8];
			System.arraycopy(plain, 0, cryptogram, 0, plain.length);
			byte[] cipherText;

			Cipher bcCipher = Cipher.getInstance("RSA/NONE/NoPadding", "BC");
			bcCipher.init(Cipher.ENCRYPT_MODE, pub);
			cipherText = bcCipher.doFinal(cryptogram);

			// Test Decipher with PKCS1 Padding
			bcCipher = Cipher.getInstance("RSA/NONE/PKCS1Padding", "BC");
			bcCipher.init(Cipher.ENCRYPT_MODE, pub);
			byte[] paddedCipherText = bcCipher.doFinal(plain);

			Cipher hsmPaddingCipher = Cipher.getInstance("RSA/ECB/PKCS1Padding", "SmartCardHSM");
			hsmPaddingCipher.init(Cipher.DECRYPT_MODE, prk);
			byte[] result = hsmPaddingCipher.doFinal(paddedCipherText);
			assert (java.util.Arrays.equals(plain, result));

			Cipher hsmCipher = Cipher.getInstance("RSA", "SmartCardHSM");

			// Test Decipher with OAEP
			bcCipher = Cipher.getInstance("RSA/NONE/OAEPWithSHA1AndMGF1Padding", "BC");
			bcCipher.init(Cipher.ENCRYPT_MODE, pub);
			byte[] oaepCipherText = bcCipher.doFinal(plain);

			Cipher hsmOAEPCipher = Cipher.getInstance("RSA/ECB/OAEPWithSHA-1AndMGF1Padding", "SmartCardHSM");
			hsmOAEPCipher.init(Cipher.DECRYPT_MODE, prk);
			result = null;
			result = hsmOAEPCipher.doFinal(oaepCipherText);
			assert (java.util.Arrays.equals(plain, result));
		} finally {
			// Delete key
			ks.deleteEntry("MasterKey");
		}
	}

	@ParameterizedTest
	@MethodSource("rsaKeyGen")
	public void wrap(SmartCardHSMRSAPrivateKeySpec spec) throws KeyStoreException, NoSuchProviderException,
	NoSuchAlgorithmException, CertificateException, IOException,
	UnrecoverableKeyException, InvalidKeySpecException,
	InvalidKeyException, SignatureException,
	InvalidAlgorithmParameterException, NoSuchPaddingException,
	IllegalBlockSizeException, BadPaddingException {

		// Generate Key Pair
		KeyPairGenerator kpg = KeyPairGenerator.getInstance("RSA", "SmartCardHSM");
		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec(
				"MasterKey", spec);
		kpg.initialize(params, null);
		KeyPair kp = kpg.generateKeyPair();

		// Get key from KeyStore
		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);
		SmartCardHSMKey prk = (SmartCardHSMKey) ks.getKey("MasterKey", null);
		PublicKey pub = ks.getCertificate("MasterKey").getPublicKey();

		try {
			// Encrypt
			byte[] plain = "Hello World".getBytes();
			byte[] cryptogram = new byte[prk.getKeySize() / 8];
			System.arraycopy(plain, 0, cryptogram, 0, plain.length);
			byte[] cipherText;

			Cipher bcCipher = Cipher.getInstance("RSA/NONE/NoPadding", "BC");
			bcCipher.init(Cipher.ENCRYPT_MODE, pub);
			cipherText = bcCipher.doFinal(cryptogram);

			Cipher hsmCipher = Cipher.getInstance("RSA", "SmartCardHSM");

			// Wrap key
			hsmCipher.init(Cipher.WRAP_MODE, prk);
			byte[] wrappedKey = hsmCipher.wrap(prk);

			// Delete key
			ks.deleteEntry("MasterKey");

			// Unwrap key
			Key k = null; // unwrap uses the key encryption key of the SmartCard-HSM
			hsmCipher.init(Cipher.UNWRAP_MODE, k);
			String wrappedKeyAlgorithm = "SmartCardHSM"; // wrappedKeyAlgorithm is not used by the Cipher implementation
			Key key = hsmCipher.unwrap(wrappedKey, wrappedKeyAlgorithm, Cipher.PRIVATE_KEY);
			SmartCardHSMKey hsmKey = (SmartCardHSMKey) key;

			// Decrypt with unwrapped key
			hsmCipher.init(Cipher.DECRYPT_MODE, hsmKey);
			byte[] plainText = hsmCipher.doFinal(cipherText);

			assert (java.util.Arrays.equals(cryptogram, plainText) == true);
		} finally {
			// Delete key
			ks.deleteEntry("MasterKey");
		}
	}



	@Test
	public void ecdhTest() throws NoSuchAlgorithmException,
	NoSuchProviderException, InvalidKeyException,
	InvalidAlgorithmParameterException, KeyStoreException,
	CertificateException, IOException, UnrecoverableKeyException {
		// Generate EC Key Pair
		KeyPairGenerator kpg = KeyPairGenerator.getInstance("EC", "SmartCardHSM");
		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec("EC1",	ecSpec);
		kpg.initialize(params, null);
		KeyPair kp1 = kpg.generateKeyPair();

		KeyPairGenerator bckpg = KeyPairGenerator.getInstance("EC", "BC");
		bckpg.initialize(ecparam);
		KeyPair kp2 = bckpg.generateKeyPair();
		PublicKey bcPublicKey = kp2.getPublic();
		Key bcPrivateKey = kp2.getPrivate();

		// KeyStore
		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);
		SmartCardHSMKey hsmPrivateKey = (SmartCardHSMKey) ks.getKey("EC1", null);
		PublicKey hsmPublicKey = ks.getCertificate("EC1").getPublicKey();

		// KeyAgreement
		KeyAgreement hsmKA = KeyAgreement.getInstance("ECDH", "SmartCardHSM");
		hsmKA.init(hsmPrivateKey);
		Key key1 = hsmKA.doPhase(bcPublicKey, true);
		assert (key1 != null);
		byte[] secret = hsmKA.generateSecret();
		assert (secret != null);

		// copy x coordinate
		int length = (secret.length - 1) / 2;
		byte[] hsmSecret = new byte[length];
		System.arraycopy(secret, 1, hsmSecret, 0, length);

		KeyAgreement bcKA = KeyAgreement.getInstance("ECDH", "BC");
		bcKA.init(bcPrivateKey);
		Key key2 = bcKA.doPhase(hsmPublicKey, true);
		byte[] bcSecret = bcKA.generateSecret();

		// HSM and BC shall have the same secret
		assert (java.util.Arrays.equals(hsmSecret, bcSecret));

		// Delete key
		ks.deleteEntry("EC1");
	}



	@Test
	public void ecdh3parties() throws NoSuchAlgorithmException,
	NoSuchProviderException, InvalidKeyException,
	InvalidAlgorithmParameterException, KeyStoreException,
	CertificateException, IOException, UnrecoverableKeyException {
		// Generate EC Key Pairs
		KeyPairGenerator kpg = KeyPairGenerator.getInstance("EC", "SmartCardHSM");

		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec("EC1", ecSpec);
		kpg.initialize(params, null);
		KeyPair kp = kpg.generateKeyPair();

		params = new SmartCardHSMParameterSpec("EC2", ecSpec);
		kpg.initialize(params, null);
		KeyPair kp2 = kpg.generateKeyPair();

		params = new SmartCardHSMParameterSpec("EC3", ecSpec);
		kpg.initialize(params, null);
		KeyPair kp3 = kpg.generateKeyPair();

		// Alice's keys
		Key alicePrivateKey = kp.getPrivate();
		PublicKey alicePublicKey = kp.getPublic();

		// Generate Bob's keys
		PublicKey bobPublicKey = kp2.getPublic();
		Key bobPrivateKey = kp2.getPrivate();

		// Generate Carol's keys
		PublicKey carolPublicKey = kp3.getPublic();
		Key carolPrivateKey = kp3.getPrivate();

		// KeyAgreement

		KeyAgreement bob = KeyAgreement.getInstance("ECDH", "SmartCardHSM");
		bob.init(bobPrivateKey);
		Key keyAB = bob.doPhase(alicePublicKey, false);

		// Carol's secret
		KeyAgreement carol = KeyAgreement.getInstance("ECDH", "SmartCardHSM");
		carol.init(carolPrivateKey);
		Key keyABC = carol.doPhase(keyAB, true);
		byte[] secret = carol.generateSecret();
		int length = (secret.length - 1) / 2;
		byte[] carolSecret = new byte[length];
		System.arraycopy(secret, 1, carolSecret, 0, length);

		Key keyBC = carol.doPhase(bobPublicKey, false);

		// Alice's secret
		KeyAgreement alice = KeyAgreement.getInstance("ECDH", "SmartCardHSM");
		alice.init(alicePrivateKey);
		Key keyBCA = alice.doPhase(keyBC, true);
		secret = alice.generateSecret();
		byte[] aliceSecret = new byte[length];
		System.arraycopy(secret, 1, aliceSecret, 0, length);

		Key keyCA = carol.doPhase(alicePublicKey, false);

		// Bob's secret
		Key keyCAB = bob.doPhase(keyCA, true);
		secret = bob.generateSecret();
		byte[] bobSecret = new byte[length];
		System.arraycopy(secret, 1, bobSecret, 0, length);

		// Compare secrets
		assert (java.util.Arrays.equals(aliceSecret, bobSecret));
		assert (java.util.Arrays.equals(aliceSecret, carolSecret));

		// Delete Keys
		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);
		ks.deleteEntry("EC1");
		ks.deleteEntry("EC2");
		ks.deleteEntry("EC3");
	}



	@Test
	public void aesGenerator() throws NoSuchAlgorithmException, NoSuchProviderException, InvalidAlgorithmParameterException, KeyStoreException, CertificateException, IOException {
		KeyGenerator kg = KeyGenerator.getInstance("AES", "SmartCardHSM");
		SmartCardHSMSecretKeySpec aesspec = new SmartCardHSMSecretKeySpec(128);
		aesspec.setAlgorithmList(new byte[] { 0x10, 0x11, (byte)0x92, (byte)0x94, (byte)0x99 } );
		SmartCardHSMParameterSpec spec = new SmartCardHSMParameterSpec("AES", aesspec);
		kg.init(spec, null);
		SecretKey key = kg.generateKey();

		aesspec = new SmartCardHSMSecretKeySpec(128);
		aesspec.setAlgorithmList(new byte[] { 0x10, 0x11, (byte)0x92, (byte)0x94, (byte)0x99 } );
		aesspec.setWrappingKey(key);
		spec = new SmartCardHSMParameterSpec("AESW", aesspec);
		kg.init(spec, null);
		SecretKey key2 = kg.generateKey();

		// Delete Keys
		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);
		ks.deleteEntry("AES");
		ks.deleteEntry("AESW");
	}



	@Test
	public void dynamicLabel() throws NoSuchAlgorithmException, NoSuchProviderException, InvalidAlgorithmParameterException, KeyStoreException, CertificateException, IOException {
		KeyGenerator kg = KeyGenerator.getInstance("AES", "SmartCardHSM");
		SmartCardHSMSecretKeySpec aesspec = new SmartCardHSMSecretKeySpec(128);
		aesspec.setAlgorithmList(new byte[] { 0x10, 0x11, (byte)0x92, (byte)0x94, (byte)0x99 } );
		SmartCardHSMParameterSpec spec = new SmartCardHSMParameterSpec("%t.%i(%r)", aesspec);
		kg.init(spec, null);
		SecretKey key = kg.generateKey();

		// Delete Keys
		KeyStore ks = KeyStore.getInstance("SmartCardHSMKeyStore", "SmartCardHSM");
		ks.load(null, null);
		ks.deleteEntry(((SmartCardHSMSecretKey)key).getLabel());


		KeyPairGenerator kpg = KeyPairGenerator.getInstance("EC", "SmartCardHSM");

		SmartCardHSMParameterSpec params = new SmartCardHSMParameterSpec("EC1-%i", ecSpec);
		kpg.initialize(params, null);
		KeyPair kp = kpg.generateKeyPair();

		ks.deleteEntry(((SmartCardHSMPrivateKey)kp.getPrivate()).getLabel());
	}
}