/*
 *  ---------
 * |.##> <##.|  Open Smart Card Development Platform (www.openscdp.org)
 * |#       #|
 * |#       #|  Copyright (c) 1999-2016 CardContact Systems GmbH
 * |'##> <##'|  32429 Minden, Germany (www.cardcontact.de)
 *  ---------
 *
 *  This file is part of OpenSCDP.
 *
 *  OpenSCDP is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  OpenSCDP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with OpenSCDP; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.cardcontact.smartcardhsmprovider;

import java.security.InvalidAlgorithmParameterException;
import java.security.ProviderException;
import java.security.SecureRandom;
import java.security.spec.AlgorithmParameterSpec;

import javax.crypto.KeyGeneratorSpi;
import javax.crypto.SecretKey;

import de.cardcontact.opencard.service.smartcardhsm.KeyDescription;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMCardService;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMSecretKey;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMSecretKeySpec;
import opencard.core.OpenCardException;
import opencard.core.service.CardServiceException;
import opencard.core.terminal.CardTerminalException;

public class KeyGenerator extends KeyGeneratorSpi {

	private SmartCardHSMProvider provider;
	private SmartCardHSMCardService schsm;
	private SmartCardHSMParameterSpec spec;

	/**
	 * @param provider the SmartCardHSMProvider
	 * @param algorithm this parameter is not in use
	 */
	public KeyGenerator(SmartCardHSMProvider provider, String algorithm) {
		this.provider = provider;
		this.schsm = provider.getSmartCardHSMCardService();
	}

	@Override
	protected void engineInit(SecureRandom random) {
		throw new ProviderException("engineInit(SecureRandom random) not supported");
	}

	@Override
	protected void engineInit(AlgorithmParameterSpec params, SecureRandom random)
			throws InvalidAlgorithmParameterException {
		if (random != null) {
			throw new ProviderException("Setting a random number generator is not supported");
		}
		if (!(params instanceof SmartCardHSMParameterSpec)) {
			throw new InvalidAlgorithmParameterException("Algorithm parameter must be instance of SmartCardHSMParameterSpec");
		}
		if (!(((SmartCardHSMParameterSpec)params).getKeySpec() instanceof SmartCardHSMSecretKeySpec)) {
			throw new InvalidAlgorithmParameterException("SmartCardHSMParameterSpec must contain SmartCardHSMSecretKeySpec");
		}
		spec = (SmartCardHSMParameterSpec)params;

	}

	@Override
	protected void engineInit(int keysize, SecureRandom random) {
		throw new ProviderException("engineInit(SecureRandom random) not supported");
	}

	@Override
	protected SecretKey engineGenerateKey() {

		SmartCardHSMSecretKeySpec skspec = (SmartCardHSMSecretKeySpec)spec.getKeySpec();

		byte[] cryptogram = null;
		byte ref;
		try {
			ref = schsm.determineFreeKeyId();
			cryptogram = schsm.generateKey(ref, skspec);
		} catch (OpenCardException e) {
			throw new ProviderException("Generating key failed", e);
		}

		byte[] algorithms = skspec.getAlgorithmList();

		boolean hasSP800_56C = false;

		for (int i = 0; i < algorithms.length; i++) {
			if (algorithms[i] == (byte)0x99) {
				hasSP800_56C = true;
				break;
			}
		}

		byte[] id = null;

		if (hasSP800_56C) {
			byte[] dparam = "KeyCheckValue".getBytes();
			byte[] res;
			try {
				res = schsm.deriveSymmetricKey(ref, (byte)0x99, dparam);
			} catch (CardTerminalException | CardServiceException e) {
				throw new ProviderException("Deriving KCV failed", e);
			}
			id = new byte[8];
			System.arraycopy(res, 0, id, 0, id.length);
		}

		KeyDescription kd = new KeyDescription(id, spec.getLabel(), (short) 0, KeyDescription.KeyTypes.AES);
		kd.setKeyRef(ref);

		try {
			schsm.storePRKD(ref, kd);
		} catch (OpenCardException e) {
			throw new ProviderException("Writing key description failed", e);
		}

		SmartCardHSMSecretKey keyRef = new SmartCardHSMSecretKey(ref, kd.getTranslatedLabel(), (short)skspec.getKeySize(), "AES");
		keyRef.setAlgorithms(algorithms);
		if (skspec.hasKeyDomain()) {
			keyRef.setKeyDomain(skspec.getKeyDomain());
		}

		if ((cryptogram != null) && (cryptogram.length > 0)) {
			keyRef.setWrapCryptogram(cryptogram);
		}

		if (id != null) {
			keyRef.setKeyId(id);
		}

		// Add key to the keymap
		schsm.addKeyToMap(keyRef);

		return keyRef;
	}
}
