/*
 *  ---------
 * |.##> <##.|  Open Smart Card Development Platform (www.openscdp.org)
 * |#       #|
 * |#       #|  Copyright (c) 1999-2016 CardContact Systems GmbH
 * |'##> <##'|  32429 Minden, Germany (www.cardcontact.de)
 *  ---------
 *
 *  This file is part of OpenSCDP.
 *
 *  OpenSCDP is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  OpenSCDP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with OpenSCDP; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.cardcontact.smartcardhsmprovider;

import java.security.InvalidAlgorithmParameterException;
import java.security.KeyPair;
import java.security.KeyPairGeneratorSpi;
import java.security.ProviderException;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.cert.CertificateException;
import java.security.spec.AlgorithmParameterSpec;
import java.security.spec.ECGenParameterSpec;
import java.security.spec.ECParameterSpec;
import java.util.HashMap;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.cardcontact.opencard.service.smartcardhsm.KeyDescription;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMCardService;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMECKey;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMECPrivateKeySpec;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMEntry;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMPrivateKey;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMRSAKey;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMRSAPrivateKeySpec;
import de.cardcontact.tlv.cvc.CardVerifiableCertificate;
import opencard.core.OpenCardException;
import opencard.opt.iso.fs.CardFilePath;

/**
 * This class provides a key pair generator for RSA and ECC keys.
 *
 * @author lew
 */
public class KeyPairGenerator {

	/**
	 * The abstract class Key provides attributes needed for the implementing
	 * classes RSA and ECC
	 *
	 * @author lew
	 */
	public static abstract class Key extends KeyPairGeneratorSpi {

		/**
		 * The private device authentication key reference.
		 */
		protected static byte PRK_DEV_AUT = 0x00;

		/**
		 * This object provides access to the SmartCard.
		 */
		protected SmartCardHSMCardService schsm;

		/**
		 * The unique key alias.
		 */
		protected String label;

		/**
		 * The new generated private key.
		 */
		protected SmartCardHSMPrivateKey privateKeyRef;

		/**
		 * The key algorithm.
		 */
		protected String algorithm;

		public Key(SmartCardHSMProvider provider, String algorithm) {
			this.schsm = provider.getSmartCardHSMCardService();
			this.algorithm = algorithm;
			if (!provider.isVerified()) throw new ProviderException("Login required.");
		}

		/**
		 * If there exists an entry to the given label
		 * then the id to the corresponding label will be returned
		 *
		 * Otherwise return a new id
		 * @return key id
		 * @throws OpenCardException
		 */
		protected byte determineID() throws OpenCardException {
			SmartCardHSMEntry entry = this.schsm.getSmartCardHSMEntry(label);
			if (entry == null) {
				return schsm.determineFreeKeyId();
			} else {
				throw new ProviderException("A key for label '" + label + "' does already exist");
			}
		}
	}

	/**
	 * Class for RSA key pair generation.
	 *
	 * @author lew
	 */
	public static class RSA extends Key {

		private final static Logger log = LoggerFactory.getLogger(RSA.class);

		/**
		 * This key spec contains the domain parameter.
		 */
		private SmartCardHSMRSAPrivateKeySpec spec;

		/**
		 * @param algorithm
		 *            the key algorithm
		 */
		public RSA(SmartCardHSMProvider provider, String algorithm) {
			super(provider, algorithm);
		}

		/**
		 * Generate a new RSA key pair. A previous initialization is required.
		 *
		 * @return the RSA key pair or null
		 */
		@Override
		public KeyPair generateKeyPair() {
			PublicKey publicKey = null;
			try {

				// the keyid for the new key
				byte keyid = determineID();
				log.debug("Generating key pair with key ID: " + keyid);
				byte[] encCert = schsm.generateKeyPair(keyid, spec);
				CardFilePath certPath = new CardFilePath(new byte[] { SmartCardHSMCardService.EECERTIFICATEPREFIX, keyid });
				this.schsm.write(certPath, 0, encCert);

				CardVerifiableCertificate cvc = new CardVerifiableCertificate("CVC", encCert);
				publicKey = (PublicKey) cvc.getPublicKey();

				byte[] id = cvc.getSubjectPublicKeyIdentifier();

				// 	Build Key description and update the corresponding ef

				KeyDescription prkd = new KeyDescription(id, label, (short) spec.getModulusSize(), KeyDescription.KeyTypes.RSA);
				prkd.setKeyRef(keyid);
				schsm.storePRKD(keyid, prkd);

				privateKeyRef = new SmartCardHSMRSAKey(keyid, prkd.getTranslatedLabel(), (short) spec.getModulusSize());

				if (spec.hasAlgorithmList()) {
					privateKeyRef.setAlgorithms(spec.getAlgorithmList());
				}
				if (spec.hasKeyDomain()) {
					privateKeyRef.setKeyDomain(spec.getKeyDomain());
				}
				privateKeyRef.setKeyId(id);

				// Add key to the keymap

				schsm.addKeyToMap(privateKeyRef);
				schsm.addCertToMap(cvc, true, keyid, label);

			} catch (OpenCardException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			} catch (CertificateException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}

			return new KeyPair(publicKey, privateKeyRef);
		}

		/**
		 * This method is not fully supported.
		 * Use initialize(AlgorithmParameterSpec spec, SecureRandom random)
		 */
		@Override
		public void initialize(int keysize, SecureRandom random) {
			if (random != null) {
				throw new ProviderException("Setting a random number generator is not supported");
			}

			this.spec = new SmartCardHSMRSAPrivateKeySpec(keysize);
		}

		private void initialize(int keysize) {
			initialize(keysize, null);
		}

		@Override
		public void initialize(AlgorithmParameterSpec spec, SecureRandom random)
				throws InvalidAlgorithmParameterException {
			log.trace("Entering initialize");
			if (!(spec instanceof SmartCardHSMParameterSpec)) {
				throw new InvalidAlgorithmParameterException("Spec is not an instance of SmartCardHSMPrivateKeySpec");
			}
			this.label = ((SmartCardHSMParameterSpec) spec).getLabel();
			this.spec = (SmartCardHSMRSAPrivateKeySpec) ((SmartCardHSMParameterSpec) spec).getKeySpec();
			this.spec.setAlgorithm(getAlgorithmFromString(this.algorithm));
		}

		/**
		 * Returns the encoded algorithm
		 *
		 * @param algo
		 * @return the encoded algorithm
		 */
		private byte[] getAlgorithmFromString(String algo) {
			HashMap<String, byte[]> map = new HashMap<String, byte[]>();
			map.put("PKCS1-v1-5-SHA-1", new byte[] { 0x04, 0x00, 0x7F, 0x00, 0x07, 0x02, 0x02, 0x02, 0x01, 0x01 });
			map.put("PKCS1-v1-5-SHA-256", new byte[] { 0x04, 0x00, 0x7F, 0x00, 0x07, 0x02, 0x02, 0x02, 0x01, 0x02 });
			map.put("PKCS1-PSS-SHA-1", new byte[] { 0x04, 0x00, 0x7F, 0x00, 0x07, 0x02, 0x02, 0x02, 0x01, 0x03 });
			map.put("PKCS1-PSS-SHA-256", new byte[] { 0x04, 0x00, 0x7F, 0x00, 0x07, 0x02, 0x02, 0x02, 0x01, 0x04 });

			return map.get(algo);
		}
	}

	/**
	 * Class for EC key pair generation.
	 *
	 * @author lew
	 */
	public static class EC extends Key {

		private final static Logger log = LoggerFactory.getLogger(EC.class);

		/**
		 * This key spec contains the domain parameter.
		 */
		private SmartCardHSMECPrivateKeySpec spec;

		public EC(SmartCardHSMProvider provider, String algorithm) {
			super(provider, algorithm);
		}

		@Override
		public void initialize(int keysize, SecureRandom random) {
			ECGenParameterSpec ecparam;

			switch(keysize) {
			case 192:
				ecparam = new ECGenParameterSpec("secp192r1");
				break;
			case 256:
				ecparam = new ECGenParameterSpec("secp256r1");
				break;
			default:
				throw new IllegalArgumentException("Unsupported key size " + keysize);
			}
			try	{
				initialize(ecparam, random);
			}
			catch(InvalidAlgorithmParameterException e) {
				throw new IllegalArgumentException("Invalid algorithm", e);
			}
		}

		@Override
		public void initialize(AlgorithmParameterSpec aps, SecureRandom random)
				throws InvalidAlgorithmParameterException {
			log.trace("initialize(spec, random) called...");
			if (aps instanceof SmartCardHSMParameterSpec) {
				SmartCardHSMParameterSpec scaps = (SmartCardHSMParameterSpec)aps;

				this.label = scaps.getLabel();
				this.spec = (SmartCardHSMECPrivateKeySpec)scaps.getKeySpec();
			} else if ((aps instanceof ECGenParameterSpec) || (aps instanceof ECParameterSpec)) {
				this.label = null;
				this.spec = new SmartCardHSMECPrivateKeySpec(aps);
			} else {
				throw new InvalidAlgorithmParameterException("Spec is not an instance of SmartCardHSMPrivateKeySpec, ECGenParameterSpec or ECParameterSpec");
			}
			this.spec.setAlgorithm(getAlgorithmFromString(this.algorithm));
		}

		/**
		 * Generate a new EC key pair. A previous initialization is required.
		 *
		 * @return the ECC key pair or null
		 */
		@Override
		public KeyPair generateKeyPair() {
			PublicKey publicKey = null;
			try {
				// Generate key pair

				byte keyid = determineID(); // the keyid for the new key
				log.debug("Generating key pair with key ID: " + keyid);
				byte[] encCert = schsm.generateKeyPair(keyid, spec);
				CardFilePath certPath = new CardFilePath(new byte[] { SmartCardHSMCardService.EECERTIFICATEPREFIX, keyid });
				this.schsm.write(certPath, 0, encCert);

				CardVerifiableCertificate cvc = new CardVerifiableCertificate("CVC", encCert);
				publicKey = (PublicKey) cvc.getPublicKey();

				byte[] id = cvc.getSubjectPublicKeyIdentifier();

				KeyDescription prkd = new KeyDescription(id, label, (short) spec.getKeySize(), KeyDescription.KeyTypes.EC);
				prkd.setKeyRef(keyid);
				schsm.storePRKD(keyid, prkd);

				privateKeyRef = new SmartCardHSMECKey(keyid, prkd.getTranslatedLabel(), (short) spec.getKeySize());

				if (spec.hasAlgorithmList()) {
					privateKeyRef.setAlgorithms(spec.getAlgorithmList());
				}
				if (spec.hasKeyDomain()) {
					privateKeyRef.setKeyDomain(spec.getKeyDomain());
				}
				privateKeyRef.setKeyId(id);

				// Add key to the keymap

				schsm.addKeyToMap(privateKeyRef);
				schsm.addCertToMap(cvc, true, keyid, label);

			} catch (OpenCardException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			} catch (CertificateException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			} catch (Exception e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}

			return new KeyPair(publicKey, privateKeyRef);
		}

		/**
		 * Returns the encoded algorithm
		 *
		 * @param algo
		 * @return the encoded algorithm
		 */
		private byte[] getAlgorithmFromString(String algo) {
			HashMap<String, byte[]> map = new HashMap<String, byte[]>();
			map.put("ECDSA-SHA-1", new byte[] { 0x04, 0x00, 0x7F, 0x00, 0x07, 0x02, 0x02, 0x02, 0x02, 0x01 });
			map.put("ECDSA-SHA-224", new byte[] { 0x04, 0x00, 0x7F, 0x00, 0x07, 0x02, 0x02, 0x02, 0x02, 0x02 });
			map.put("ECDSA-SHA-256", new byte[] { 0x04, 0x00, 0x7F, 0x00, 0x07, 0x02, 0x02, 0x02, 0x02, 0x03 });

			return map.get(algo);
		}

	}
}
