/*
 *  ---------
 * |.##> <##.|  Open Smart Card Development Platform (www.openscdp.org)
 * |#       #|
 * |#       #|  Copyright (c) 1999-2016 CardContact Systems GmbH
 * |'##> <##'|  32429 Minden, Germany (www.cardcontact.de)
 *  ---------
 *
 *  This file is part of OpenSCDP.
 *
 *  OpenSCDP is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  OpenSCDP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with OpenSCDP; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.cardcontact.smartcardhsmprovider;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.security.AlgorithmParameters;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;
import java.security.ProviderException;
import java.security.SecureRandom;
import java.security.spec.AlgorithmParameterSpec;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.CipherSpi;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.NoSuchPaddingException;
import javax.crypto.ShortBufferException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.cardcontact.opencard.service.smartcardhsm.DecipherCardService;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMCardService;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMKey;
import de.cardcontact.tlv.ConstructedTLV;
import de.cardcontact.tlv.PrimitiveTLV;
import de.cardcontact.tlv.TLVEncodingException;
import de.cardcontact.tlv.Tag;
import opencard.core.service.CardServiceException;
import opencard.core.terminal.CardTerminalException;
import opencard.opt.iso.fs.CardFilePath;
import opencard.opt.iso.fs.FileAccessCardService;

/**
 * This class implements RSA decryption functionality.
 * Encryption, wrapping and unwrapping aren't supported.
 *
 * @author lew
 *
 */
public class SmartCardHSMCipher extends CipherSpi {

	private final static Logger log = LoggerFactory.getLogger(SmartCardHSMCipher.class);

	/**
	 * This object provides access to the SmartCard.
	 */
	private SmartCardHSMCardService schsm;

	/**
	 * Must be one of
	 * ENCRYPT_MODE,
	 * DECRYPT_MODE,
	 * WRAP_MODE,
	 * UNWRAP_MODE
	 */
	private int selectedMode;

	/**
	 * algorithm/mode/padding
	 * or algorithm
	 */
	private String algorithm;

	/**
	 * The the input for the decipher operation
	 */
	private ByteArrayOutputStream input;

	/**
	 * Reference to the SmartCardHSM RSA private Key used for decryption
	 */
	private SmartCardHSMKey key;
	private SmartCardHSMProvider provider;

	public SmartCardHSMCipher(SmartCardHSMProvider provider, String algorithm) {
		this.provider = provider;
	    this.schsm = provider.getSmartCardHSMCardService();
	    this.algorithm = algorithm;

	    if (!provider.isVerified()) throw new ProviderException("Login required.");
	}

	protected byte[] decipher() {
		byte algorithmID;
		switch (algorithm) {
		case "RSA/None/PKCS1Padding":
		case "RSA/ECB/PKCS1Padding":
			algorithmID = DecipherCardService.RSA_DECRYPTION_V15;
			break;
		case "RSA/ECB/OAEPWithSHA-1AndMGF1Padding":
			algorithmID = DecipherCardService.RSA_DECRYPTION_OAEP;
			break;
		default:
			algorithmID = DecipherCardService.RSA_DECRYPTION_PLAIN;
		}

		byte[] result;
		try {
			result = schsm.decipher(key, input.toByteArray(), algorithmID);
		} catch (CardTerminalException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (CardServiceException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}

		return result;
	}

	@Override
	protected byte[] engineDoFinal(byte[] input, int inputOffset, int inputLen)
			throws IllegalBlockSizeException, BadPaddingException {

		if (input != null) {
			if ((this.input.size() + inputLen) > (key.getKeySize() / 8)) {
				throw new ProviderException("Input Data exceeds the lenght of the modulus");
			}

			this.input.write(input, inputOffset, inputLen);
		}

		byte[] res = decipher();

		selectedMode = -1;
		this.input.reset();
		this.input = null;

		return res;
	}

	@Override
	protected int engineDoFinal(byte[] input, int inputOffset, int inputLen, byte[] output,
			int outputOffset) throws ShortBufferException, IllegalBlockSizeException,
			BadPaddingException {

		byte[] res = engineDoFinal(input, inputOffset, inputLen);

		if (res.length > (output.length - outputOffset)) {
			throw new ShortBufferException("The output buffer is too short");
		}

		System.arraycopy(res, 0, output, outputOffset, res.length);

		return res.length;
	}

	@Override
	protected int engineGetBlockSize() {
		throw new UnsupportedOperationException();
	}

	@Override
	protected byte[] engineGetIV() {
		throw new UnsupportedOperationException();
	}

	@Override
	protected int engineGetOutputSize(int arg0) {
		throw new UnsupportedOperationException();
	}

	@Override
	protected AlgorithmParameters engineGetParameters() {
		return null;
	}

	/**
	 * @param mode Must be Cipher.DECRYPT_MODE
	 * @param key Key must be type of SmartCardHSMRSAKey
	 * @param rnd not necessary
	 */
	@Override
	protected void engineInit(int mode, Key key, SecureRandom rnd)
			throws InvalidKeyException {

		if (mode != Cipher.DECRYPT_MODE
				&& mode != Cipher.WRAP_MODE
				&& mode != Cipher.UNWRAP_MODE) {
			throw new UnsupportedOperationException("Only Decryption, Wrap and Unwrap mode are supported");
		}

		selectedMode = mode;

		// If the selected mode is WRAP or UNWRAP
		// the SmartCardHSM uses Device Key Encryption Key and NOT the key parameter.
		// Otherwise the key type must be SmartCardHSMKey
		if(mode != Cipher.WRAP_MODE && mode != Cipher.UNWRAP_MODE) {

			if (!(key instanceof SmartCardHSMKey)) {
				throw new InvalidKeyException("Key must be type of SmartCardHSMKey");
			}

			this.key = (SmartCardHSMKey) key;

			input = new ByteArrayOutputStream();
		}
	}

	@Override
	protected void engineInit(int arg0, Key arg1, AlgorithmParameterSpec arg2,
			SecureRandom arg3) throws InvalidKeyException,
			InvalidAlgorithmParameterException {
		throw new UnsupportedOperationException();
	}

	@Override
	protected void engineInit(int arg0, Key arg1, AlgorithmParameters arg2,
			SecureRandom arg3) throws InvalidKeyException,
			InvalidAlgorithmParameterException {
		throw new UnsupportedOperationException();
	}

	@Override
	protected void engineSetMode(String arg0) throws NoSuchAlgorithmException {
		throw new UnsupportedOperationException();
	}

	@Override
	protected void engineSetPadding(String arg0) throws NoSuchPaddingException {
		throw new UnsupportedOperationException();
	}

	@Override
	protected byte[] engineUpdate(byte[] input, int inputOffset, int inputLen) {

		if ((this.input.size() + inputLen) > key.getKeySize()) {
			throw new ProviderException("Input Data exceeds the lenght of the modulus");
		}

		this.input.write(input, inputOffset, inputLen);

		return null;
	}

	@Override
	protected int engineUpdate(byte[] input, int inputOffset, int inputLen, byte[] output,
			int outputOffset) throws ShortBufferException {

		engineUpdate(input, inputOffset, inputLen);

		return 0;
	}

	@Override
	protected int engineGetKeySize(Key key) throws InvalidKeyException {
		if (key instanceof SmartCardHSMKey) {
			key = (SmartCardHSMKey)key;
			return ((SmartCardHSMKey) key).getKeySize();
		}
		throw new InvalidKeyException("Key must be type of SmartCardHSMKey");
	}

	@Override
	protected byte[] engineWrap(Key key) throws InvalidKeyException {
		if (selectedMode != Cipher.WRAP_MODE) {
			throw new IllegalStateException("Cipher is not in wrap mode");
		}

		if (!(key instanceof SmartCardHSMKey)) {
			throw new InvalidKeyException("Key must be type of SmartCardHSMKey");
		}

		byte[] wrappedKeyBinary;
		byte[] prkdBinary;
		byte[] certBinary;
		ConstructedTLV sequence;
		SmartCardHSMKey hsmKey = (SmartCardHSMKey) key;

		try {
			// First wrap specified key

			wrappedKeyBinary = schsm.wrapKey(hsmKey.getKeyRef());

			// Get the corresponding private key description

			CardFilePath path = new CardFilePath(new byte[] { SmartCardHSMCardService.PRKDPREFIX, hsmKey.getKeyRef() });
			prkdBinary = schsm.read(path, 0, FileAccessCardService.READ_SEVERAL);

			// Get the corresponding certificate
			path = new CardFilePath(new byte[] { SmartCardHSMCardService.EECERTIFICATEPREFIX, hsmKey.getKeyRef() });
			certBinary = schsm.read(path, 0, FileAccessCardService.READ_SEVERAL);
		} catch (CardServiceException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (CardTerminalException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}

		// Build TLV sequence
		try {
			sequence = new ConstructedTLV(Tag.SEQUENCE | Tag.CONSTRUCTED);
			PrimitiveTLV wrappedKey = new PrimitiveTLV(Tag.OCTET_STRING, wrappedKeyBinary);
			PrimitiveTLV prkd = new PrimitiveTLV(prkdBinary);
			PrimitiveTLV cert = new PrimitiveTLV(certBinary);
			sequence.add(wrappedKey);
			sequence.add(prkd);
			sequence.add(cert);
		} catch (TLVEncodingException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}

		return sequence.getBytes();
	}

	@Override
	protected Key engineUnwrap(byte[] wrappedKey, String wrappedKeyAlgorithm, int wrappedKeyType) {
		if (selectedMode != Cipher.UNWRAP_MODE) {
			throw new IllegalStateException("Cipher is not in unwrap mode");
		}
		byte[] key = null;
		byte[] cert = null;
		byte[] desc = null;
		byte kid;

		try {
			ConstructedTLV sequence = new ConstructedTLV(wrappedKey);

			key = sequence.get(0).getValue();
			desc = sequence.get(1).getBytes();
			cert = sequence.get(2).getValue();

			kid = schsm.determineFreeKeyId();
			schsm.unwrapKey(kid, key);

			CardFilePath path = new CardFilePath(new byte[] { SmartCardHSMCardService.PRKDPREFIX, kid });
			schsm.write(path, 0, desc);

			path = new CardFilePath(new byte[] { SmartCardHSMCardService.EECERTIFICATEPREFIX, kid });
			schsm.write(path, 0, cert);

			return schsm.addKey(kid);
		} catch (CardTerminalException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (CardServiceException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (IOException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (TLVEncodingException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}
	}
}
