/*
 *  ---------
 * |.##> <##.|  Open Smart Card Development Platform (www.openscdp.org)
 * |#       #|
 * |#       #|  Copyright (c) 1999-2016 CardContact Systems GmbH
 * |'##> <##'|  32429 Minden, Germany (www.cardcontact.de)
 *  ---------
 *
 *  This file is part of OpenSCDP.
 *
 *  OpenSCDP is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  OpenSCDP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with OpenSCDP; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.cardcontact.smartcardhsmprovider;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.KeyFactory;
import java.security.NoSuchAlgorithmException;
import java.security.ProviderException;
import java.security.SecureRandom;
import java.security.interfaces.ECPublicKey;
import java.security.spec.AlgorithmParameterSpec;
import java.security.spec.ECField;
import java.security.spec.ECFieldFp;
import java.security.spec.ECParameterSpec;
import java.security.spec.ECPoint;
import java.security.spec.ECPublicKeySpec;
import java.security.spec.EllipticCurve;
import java.security.spec.InvalidKeySpecException;

import javax.crypto.KeyAgreementSpi;
import javax.crypto.SecretKey;
import javax.crypto.ShortBufferException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMCardService;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMECKey;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMKey;
import opencard.core.service.CardServiceException;
import opencard.core.terminal.CardTerminalException;

/**
 * Class for ECDH Key Agreement
 *
 * @author lew
 *
 */
public class SmartCardHSMKeyAgreement extends KeyAgreementSpi {

	private final static Logger log = LoggerFactory.getLogger(SmartCardHSMKeyAgreement.class);

	/**
	 * This object provides access to the SmartCard.
	 */
	private SmartCardHSMCardService schsm;

	/**
	 * Reference to the SmartCardHSM private key
	 */
	private SmartCardHSMKey privateKey;

	/**
	 * EC parameter for key generation in doPhase
	 */
	private ECParameterSpec ecparam;

	/**
	 * Indicates that key agreement has finished
	 * and the shared secret can be obtained.
	 */
	private boolean finishedKeyAgreement;

	/**
	 * The shared secret consisting of
	 * '04' || x || y point coordinates
	 * on EC curve
	 */
	private byte[] secret;

	/**
     * @param provider the SmartCardHSMProvider
     * @param algorithm this parameter is not in use
     */
	public SmartCardHSMKeyAgreement(SmartCardHSMProvider provider, String algorithm) {
	    this.schsm = provider.getSmartCardHSMCardService();

	    if (!provider.isVerified()) throw new ProviderException("Login required.");
	}

	@Override
	protected Key engineDoPhase(Key key, boolean lastPhase)
			throws InvalidKeyException, IllegalStateException {

		byte[] encodedECP = null;
		try {
			encodedECP = schsm.performECCDH(privateKey, this.getPublicKeyComponents(key));
		} catch (CardServiceException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (CardTerminalException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}
		int size = (encodedECP.length - 1) / 2;
		byte[] x = new byte[size];
		byte[] y = new byte[size];
		System.arraycopy(encodedECP, 1, x, 0, size);
		System.arraycopy(encodedECP, 1 + size, y, 0, size);
		ECPoint ecp = new ECPoint(new BigInteger(x), new BigInteger(y));

		Key newKey = null;
		try {
			ECPublicKeySpec spec = new ECPublicKeySpec(ecp, ecparam);
			KeyFactory kf = KeyFactory.getInstance("EC");
			newKey = kf.generatePublic(spec);
		} catch (NoSuchAlgorithmException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (InvalidKeySpecException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}

		if (lastPhase) {
			finishedKeyAgreement = true;
			secret = encodedECP;
		}

		return newKey;
	}

	/**
	 * @return Concatenation of '04' || x || y point coordinates on EC curve
	 */
	@Override
	protected byte[] engineGenerateSecret() throws IllegalStateException {
		if (!finishedKeyAgreement) {
			throw new IllegalStateException("Can't genereate secret when this isn't the last phase.");
		}

		return secret;
	}

	@Override
	protected SecretKey engineGenerateSecret(String algorithm)
			throws IllegalStateException, NoSuchAlgorithmException,
			InvalidKeyException {
		throw new UnsupportedOperationException();
	}

	@Override
	protected int engineGenerateSecret(byte[] sharedSecret, int offset)
			throws IllegalStateException, ShortBufferException {

		if (!finishedKeyAgreement) {
			throw new IllegalStateException("Can't genereate secret when this isn't the last phase.");
		}

		if (secret.length + offset > sharedSecret.length) {
			throw new ShortBufferException("The input buffer is too short");
		}

		System.arraycopy(secret, 0, sharedSecret, offset, secret.length);

		return sharedSecret.length;
	}

	@Override
	protected void engineInit(Key key, SecureRandom rnd)
			throws InvalidKeyException {

		if (key instanceof SmartCardHSMKey) {
			privateKey = (SmartCardHSMKey) key;
		} else {
			throw new InvalidKeyException("Key must be type of SmartCardHSMECKey");
		}

		// create default ECParamSpec

		BigInteger prime = new BigInteger("A9FB57DBA1EEA9BC3E660A909D838D726E3BF623D52620282013481D1F6E5377", 16);
		ECField field = new ECFieldFp(prime);
		BigInteger a, b;
		a = new BigInteger("7D5A0975FC2C3057EEF67530417AFFE7FB8055C126DC5C6CE94A4B44F330B5D9", 16);
		b = new BigInteger("26DC5C6CE94A4B44F330B5D9BBD77CBF958416295CF7E1CE6BCCDC18FF8C07B6", 16);
		EllipticCurve curve = new EllipticCurve(field, a, b);

		BigInteger x = new BigInteger("8BD2AEB9CB7E57CB2C4B482FFC81B7AFB9DE27E1E3BD23C23A4453BD9ACE3262", 16);
		BigInteger y = new BigInteger("547EF835C3DAC4FD97F8461A14611DC9C27745132DED8E545C1D54C72F046997", 16);
		ECPoint g = new ECPoint(x, y);
		BigInteger n = new BigInteger("A9FB57DBA1EEA9BC3E660A909D838D718C397AA3B561A6F7901E0E82974856A7", 16);
		int h = 1;
		ecparam = new ECParameterSpec(curve, g, n, h);
	}

	@Override
	protected void engineInit(Key key, AlgorithmParameterSpec param,
			SecureRandom rnd) throws InvalidKeyException,
			InvalidAlgorithmParameterException {

		if (key instanceof SmartCardHSMECKey) {
			privateKey = (SmartCardHSMECKey) key;
		} else {
			throw new InvalidKeyException("Key must be type of SmartCardHSMECKey");
		}

		if (param instanceof ECParameterSpec) {
			ecparam = (ECParameterSpec) param;
		} else {
			throw new InvalidAlgorithmParameterException("Parameter must be type of ECParameterSpec");
		}
	}

	/**
	 * Extract from given EC key x and y coordinates
	 * and return the encoded coordinates.
	 *
	 * @param key the EC key
	 * @return concatenation of '04' || x || y point coordinates
	 * @throws InvalidKeyException
	 */
	private byte[] getPublicKeyComponents(Key key) throws InvalidKeyException {
		if (key instanceof ECPublicKey) {
			ECPublicKey ecKey = (ECPublicKey)key;
			ECPoint p = ecKey.getW();
			BigInteger x = p.getAffineX();
			BigInteger y = p.getAffineY();

			ByteArrayOutputStream basePointG = new ByteArrayOutputStream();
			basePointG.write(0x04);
			try {
				basePointG.write(unsignedBigIntegerToByteArray(x, 256));
				basePointG.write(unsignedBigIntegerToByteArray(y, 256));

			} catch (IOException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}
			return basePointG.toByteArray();
		} else {
			throw new InvalidKeyException("Key must be type of ECPublicKey");
		}
	}

	/**
	 * Convert unsigned big integer into byte array, stripping of a
	 * leading 00 byte
	 *
	 * This conversion is required, because the Java BigInteger is a signed
	 * value, whereas the byte arrays containing key components are unsigned by default
	 *
	 * @param bi    BigInteger value to be converted
	 * @param size  Number of bits
	 * @return      Byte array containing unsigned integer value
	 */
	protected static byte[] unsignedBigIntegerToByteArray(BigInteger bi, int size) {
		byte[] s = bi.toByteArray();
		size = (size >> 3) + ((size & 0x7) == 0 ? 0 : 1);
		byte[] d = new byte[size];
		int od = size - s.length;
		int os = 0;
		if (od < 0) {  // Number is longer than expected
			if ((od < -1) || s[0] != 0) {   // If it is just a leading zero, then we cut it off
				throw new IllegalArgumentException("Size mismatch converting big integer to byte array");
			}
			os = -od;
			od = 0;
		}
		size = size - od;

		System.arraycopy(s, os, d, od, size);
		return d;
	}
}
