/*
 *  ---------
 * |.##> <##.|  Open Smart Card Development Platform (www.openscdp.org)
 * |#       #|
 * |#       #|  Copyright (c) 1999-2016 CardContact Systems GmbH
 * |'##> <##'|  32429 Minden, Germany (www.cardcontact.de)
 *  ---------
 *
 *  This file is part of OpenSCDP.
 *
 *  OpenSCDP is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  OpenSCDP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with OpenSCDP; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.cardcontact.smartcardhsmprovider;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.Key;
import java.security.KeyStoreException;
import java.security.KeyStoreSpi;
import java.security.NoSuchAlgorithmException;
import java.security.ProviderException;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.util.Date;
import java.util.Enumeration;
import java.util.Vector;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.cardcontact.opencard.service.smartcardhsm.CertificateDescription;
import de.cardcontact.opencard.service.smartcardhsm.KeyDescription;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMCardService;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMEntry;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMKey;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMRSAKey;
import de.cardcontact.tlv.TLVEncodingException;
import opencard.core.OpenCardException;
import opencard.core.service.CardServiceException;
import opencard.core.terminal.CardTerminalException;
import opencard.core.util.HexString;
import opencard.opt.iso.fs.CardFilePath;
import opencard.opt.iso.fs.CardIOException;
import opencard.opt.service.CardServiceResourceNotFoundException;

/**
 * This class represents a storage for cryptographic keys and certificates.
 *
 * @author lew
 */
public class SmartCardHSMKeyStore extends KeyStoreSpi {

	private final static Logger log = LoggerFactory.getLogger(SmartCardHSMKeyStore.class);

	private SmartCardHSMProvider provider;

	/**
	 * An Enumeration of all alias names of keys and certificates stored in the
	 * key store.
	 */
	private Vector<String> aliases;

	public SmartCardHSMKeyStore(SmartCardHSMProvider provider, String algorithm) {
		this.provider = provider;
	}

	@Override
	public Enumeration<String> engineAliases() {
		SmartCardHSMCardService schsm = provider.getSmartCardHSMCardService();
		try {
			aliases = schsm.getAliases();
		} catch (OpenCardException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (CertificateException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (TLVEncodingException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}
		return aliases.elements();
	}

	@Override
	public boolean engineContainsAlias(String alias) {
		SmartCardHSMCardService schsm = provider.getSmartCardHSMCardService();
		return schsm.containsLabel(alias);
	}

	@Override
	public void engineDeleteEntry(String alias) throws KeyStoreException {
		if (!provider.isVerified()) throw new ProviderException("Login required.");

		SmartCardHSMCardService schsm = provider.getSmartCardHSMCardService();
		try {
			schsm.removeEntry(alias);
		} catch (CardServiceResourceNotFoundException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new KeyStoreException(alias + " not found.");
		} catch (CardTerminalException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (CardIOException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (CardServiceException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}
	}

	@Override
	public java.security.cert.Certificate engineGetCertificate(String alias) {
		SmartCardHSMCardService schsm = provider.getSmartCardHSMCardService();
		SmartCardHSMEntry entry = schsm.getSmartCardHSMEntry(alias);
		if (entry == null) {
			return null;
		}
		return entry.getCert();
	}

	@Override
	public String engineGetCertificateAlias(Certificate arg0) {
		throw new UnsupportedOperationException();
	}

	/**
	 * Returns a certificate array containing the certificate corresponding to
	 * the alias
	 */
	@Override
	public Certificate[] engineGetCertificateChain(String alias) {
		return new Certificate[] { engineGetCertificate(alias) };
	}

	@Override
	public Date engineGetCreationDate(String alias) {
		return new Date();
	}

	/**
	 * @return the private SmartCardHSMKey
	 */
	@Override
	public Key engineGetKey(String alias, char[] password) throws NoSuchAlgorithmException, UnrecoverableKeyException {
		SmartCardHSMCardService schsm = provider.getSmartCardHSMCardService();
		SmartCardHSMEntry entry = null;
		if ((password != null) && (password.length > 0)) {
			String pw = "";
			for (int i = 0; i < password.length; i++) {
				pw += password[i];
			}
			boolean verified;
			try {
				verified = schsm.verifyPassword(null, 0, pw.getBytes());
			} catch (CardServiceException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			} catch (CardTerminalException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}
		}

		entry = schsm.getSmartCardHSMEntry(alias);

		if (entry == null) {
			throw new UnrecoverableKeyException("No key found.");
		}

		return entry.getKey();
	}

	@Override
	public boolean engineIsCertificateEntry(String alias) {
		SmartCardHSMCardService schsm = provider.getSmartCardHSMCardService();
		SmartCardHSMEntry entry = schsm.getSmartCardHSMEntry(alias);
		if (entry == null) {
			return false;
		}
		return !entry.isEECertificate();
	}

	@Override
	public boolean engineIsKeyEntry(String alias) {
		SmartCardHSMCardService schsm = provider.getSmartCardHSMCardService();
		SmartCardHSMEntry entry = schsm.getSmartCardHSMEntry(alias);
		if (entry == null) {
			return false;
		}
		return entry.isKeyEntry();
	}

	@Override
	public void engineLoad(InputStream stream, char[] password) throws IOException, NoSuchAlgorithmException,
	CertificateException {

		SmartCardHSMCardService schsm = provider.getSmartCardHSMCardService();
		if (password != null) {
			String pwd = new String(password);
			boolean verified = schsm.verifyPassword(null, 0, pwd.getBytes());
			if (!verified) {
				log.error("Login failed. Wrong PIN?");
				throw new IOException("Login failed. Wrong PIN?");
			}
		}

		try {
			aliases = schsm.getAliases();
		} catch (TLVEncodingException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}
	}

	/**
	 * @throws ProviderException if no provider login has been performed
	 */
	@Override
	public void engineSetCertificateEntry(String alias, Certificate cert) throws KeyStoreException {
		SmartCardHSMCardService schsm = provider.getSmartCardHSMCardService();
		SmartCardHSMEntry entry = schsm.getSmartCardHSMEntry(alias);
		byte id;

		if (!provider.isVerified()) throw new ProviderException("Login required.");

		try {
			if (entry != null) {
				SmartCardHSMKey key = entry.getKey();
				if (key != null) {
					String path = ":CE" + HexString.hexify(key.getKeyRef());
					CardFilePath file = new CardFilePath(path);

					schsm.write(file, 0, cert.getEncoded());
					schsm.addCertToMap(cert, true, key.getKeyRef(), alias);
				} else {
					id = entry.getId();

					// Store certificate
					String path = ":CA" + HexString.hexify(id);
					CardFilePath file = new CardFilePath(path);
					schsm.write(file, 0, cert.getEncoded());
					schsm.addCertToMap(cert, false, id, alias);
				}
			} else {

				id = schsm.determineFreeCAId();
				if (id == -1) throw new KeyStoreException("Storage limit for CA Certificates reached.");

				// Store certificate
				String path = ":CA" + HexString.hexify(id);
				CardFilePath file = new CardFilePath(path);
				schsm.write(file, 0, cert.getEncoded());

				// Store alias
				path = ":C8" + HexString.hexify(id);
				file = new CardFilePath(path);
				schsm.write(file, 0, CertificateDescription.buildCertDescription(alias, cert.getPublicKey(), new byte[]{(byte) 0xCA, id}));

				schsm.addCertToMap(cert, false, id, alias);
			}
		} catch (OpenCardException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (CertificateException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (TLVEncodingException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}
	}

	@Override
	public void engineSetKeyEntry(String alias, byte[] key, Certificate[] chain) throws KeyStoreException {
		throw new UnsupportedOperationException();
	}

	@Override
	public void engineSetKeyEntry(String alias, Key key, char[] password, Certificate[] chain) throws KeyStoreException {
		SmartCardHSMKey skey = (SmartCardHSMKey)key;
		SmartCardHSMCardService schsm = provider.getSmartCardHSMCardService();

		if (!provider.isVerified()) throw new ProviderException("Login required.");

		String oldlabel = skey.getLabel();
		skey.setLabel(alias);

		try	{
			KeyDescription prkd;
			byte[] id = new byte[] { skey.getKeyRef() };
			if (key instanceof SmartCardHSMRSAKey) {
				prkd = new KeyDescription(id, alias, (short) skey.getKeySize(), KeyDescription.KeyTypes.RSA);
			} else {
				prkd = new KeyDescription(id, alias, (short) skey.getKeySize(), KeyDescription.KeyTypes.EC);
			}
			schsm.storePRKD(skey.getKeyRef(), prkd);

			if (oldlabel != null) {
				schsm.renameEntry(oldlabel, alias);
			} else {
				schsm.addKeyToMap(skey);
			}

			if (chain.length >= 1) {
				String path = ":CE" + HexString.hexify(skey.getKeyRef());
				CardFilePath file = new CardFilePath(path);

				schsm.write(file, 0, chain[0].getEncoded());
				schsm.addCertToMap(chain[0], true, skey.getKeyRef(), alias);
			}
		} catch (CardServiceException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (CardTerminalException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (CertificateEncodingException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}
	}

	@Override
	public int engineSize() {
		SmartCardHSMCardService schsm = provider.getSmartCardHSMCardService();
		try	{
			Vector<String> aliases = schsm.getAliases();
			return aliases.size();
		}
		catch(Exception e) {
			log.error(e.getLocalizedMessage(), e);
			return 0;
		}
	}

	@Override
	public void engineStore(OutputStream arg0, char[] arg1) throws IOException, NoSuchAlgorithmException,
	CertificateException {
		return;
	}
}
