/*
 *  ---------
 * |.##> <##.|  Open Smart Card Development Platform (www.openscdp.org)
 * |#       #|
 * |#       #|  Copyright (c) 1999-2016 CardContact Systems GmbH
 * |'##> <##'|  32429 Minden, Germany (www.cardcontact.de)
 *  ---------
 *
 *  This file is part of OpenSCDP.
 *
 *  OpenSCDP is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  OpenSCDP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with OpenSCDP; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.cardcontact.smartcardhsmprovider;

import java.io.IOException;
import java.lang.reflect.Constructor;
import java.security.AuthProvider;
import java.security.Provider;
import java.security.ProviderException;
import java.security.Security;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.callback.UnsupportedCallbackException;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.cardcontact.opencard.factory.GlobalPlatformCardServiceFactory;
import de.cardcontact.opencard.factory.IsoCardServiceFactory;
import de.cardcontact.opencard.factory.RemoteClientCardServiceFactory;
import de.cardcontact.opencard.factory.SmartCardHSMCardServiceFactory;
import de.cardcontact.opencard.service.StatusWordTable;
import de.cardcontact.opencard.service.remoteclient.RemoteClient;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMCardService;
import de.cardcontact.opencard.terminal.smartcardio.SmartCardIOFactory;
import de.cardcontact.tlv.HexString;
import opencard.core.OpenCardException;
import opencard.core.event.CTListener;
import opencard.core.event.CardTerminalEvent;
import opencard.core.event.EventGenerator;
import opencard.core.service.CardRequest;
import opencard.core.service.CardServiceException;
import opencard.core.service.CardServiceFactory;
import opencard.core.service.CardServiceRegistry;
import opencard.core.service.SmartCard;
import opencard.core.terminal.CardID;
import opencard.core.terminal.CardTerminal;
import opencard.core.terminal.CardTerminalException;
import opencard.core.terminal.CardTerminalFactory;
import opencard.core.terminal.CardTerminalRegistry;
import opencard.core.terminal.CommandAPDU;
import opencard.core.terminal.ResponseAPDU;
import opencard.core.terminal.SlotChannel;
import opencard.core.util.APDUFormatter;
import opencard.core.util.APDUTracer;
import opencard.core.util.OpenCardPropertyLoadingException;
import opencard.opt.util.PassThruCardServiceFactory;

/**
 * SmartCardHSM Cryptographic Service Provider *
 *
 * @author lew
 *
 */
public final class SmartCardHSMProvider extends AuthProvider implements CTListener, APDUTracer {

	private static final long serialVersionUID = 4737690040987973156L;

	private final static Logger log = LoggerFactory.getLogger(SmartCardHSMProvider.class);

	private static String PROVIDER_NAME = "SmartCardHSM";

	private static double VERSION = 1.0;

	private static String INFO = "SmartCardHSM provider implementing x";

	private boolean doSecureMessaging = true;

	private int doOCFShutdown = 0;

	private boolean usePinPad = false;

	private boolean usePKACallback = false;

	private SmartCard sc = null;

	private String terminalName = null;

	private CardTerminal ct = null;

	private int slotID = 0;

	private SmartCardHSMCardService schsm;

	private PasswordCallback passwordCallback;

	private PublicKeyAuthenticationCallback pkaCallback;

	private CallbackHandler callBackHandler = null;

	public SmartCardHSMProvider() {

		super(PROVIDER_NAME, VERSION, INFO);

		log.debug("Super constructor called...");

		initProviderProperties();

		try {
			register();
		} catch (Exception e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}
	}

	public SmartCardHSMProvider(String name, boolean handleEvents) {

		super(PROVIDER_NAME + "/" + name, VERSION, INFO);

		log.debug("Super constructor called...");

		this.terminalName = name;

		initProviderProperties();

		if (handleEvents) {
			try {
				register();
			} catch (Exception e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}
		}
	}



	public SmartCardHSMProvider(String name) {
		this(name, true);
	}



	/**
	 * Clean up and remove all SmartCard-HSM security provider
	 */
	public static void removeProviders() {
		Provider[] providers = Security.getProviders();
		for (Provider provider : providers) {
			if (provider.getName().startsWith(PROVIDER_NAME)) {
				SmartCardHSMProvider scprov = (SmartCardHSMProvider)provider;
				scprov.unregister();
				Security.removeProvider(scprov.getName());
			}
		}
	}



	/**
	 * @return true if the user is authenticated, false otherwise
	 */
	public boolean isVerified() {
		boolean status = false;

		if (schsm == null) {
			return false;
		}

		try {
			status = schsm.getSecurityStatus();
		} catch (CardServiceException e) {
			// Ignore

		} catch (CardTerminalException e) {
			// Ignore
		}

		if (!status) {
			status = verify();
		}

		return status;
	}



	private boolean verify() {
		if (schsm == null) {
			throw new ProviderException("CardService not initialized");
		}

		if (callBackHandler != null) {
			try {
				log.debug("Get password callback from application");
				if (this.usePKACallback) {
					callBackHandler.handle(new Callback[] { passwordCallback, pkaCallback });
				} else {
					callBackHandler.handle(new Callback[] { passwordCallback });
				}
			} catch (IOException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			} catch (UnsupportedCallbackException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}

			if (passwordCallback.getPassword() != null) {
				schsm.setCHVDialog(new OCFCallback(passwordCallback));
				try {
					return schsm.verifyPassword();
				} catch (OpenCardException e) {
					log.error(e.getLocalizedMessage(), e);
					throw new ProviderException(e);
				}
			} else if (pkaCallback.getUrl() != null) {
				try {
					RemoteClient rc = (RemoteClient)sc.getCardService(RemoteClient.class, true);
					rc.update(pkaCallback.getUrl(), null, null);

					return schsm.getSecurityStatus();
				} catch (Exception e) {
					log.error(e.getLocalizedMessage(), e);
					throw new ProviderException(e);
				}
			} else {
				throw new ProviderException("Neither a password nor an URL to an authentication server was configured");
			}
		} else {
			try {
				return schsm.verifyPassword();
			} catch (OpenCardException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}
		}
	}



	/**
	 * Enable or disable secure messaging.
	 * By default secure messaging is enabled.
	 * @param doSecureMessaging
	 */
	public void setSecureMessaging(boolean doSecureMessaging) {
		this.doSecureMessaging = doSecureMessaging;
	}


	/**
	 * Enable or disable pin pad.
	 * By default the pin pad is disabled.
	 * @param usePinPad true if pin pad shall be used
	 */
	public void useTerminalPinPad(boolean usePinPad) {
		this.usePinPad = usePinPad;

	}



	/**
	 * Enable or disable using the Public Key Authentication Callback.
	 * By default the callback is disabled.
	 * @param usePinPad true if pin pad shall be used
	 */
	public void usePKACallback(boolean usePKACallback) {
		this.usePKACallback = usePKACallback;

	}

	/**
	 * Add all implemented services to the provider
	 */
	private void initProviderProperties() {
		passwordCallback = new PasswordCallback("SmartCardHSMPasswordCallback", false);
		pkaCallback = new PublicKeyAuthenticationCallback();

		log.debug("Initializing provider algorithms...");

		Vector<String> alias;

		// SecureRandom

		putService(new SCHSMService(this, "SecureRandom", "NativePRNG",
				"de.cardcontact.smartcardhsmprovider.SecureRandom"));

		// Key Generator
		putService(new SCHSMService(this, "KeyGenerator", "AES",
				"de.cardcontact.smartcardhsmprovider.KeyGenerator"));

		// KeyPairGenerator

		alias = new Vector<String>();
		alias.addElement("RSA");
		putService(new KeyPairGenService(this, "KeyPairGenerator", "RSA//PKCS1-v1-5-SHA-1",
				"de.cardcontact.smartcardhsmprovider.KeyPairGenerator$RSA", alias));

		putService(new KeyPairGenService(this, "KeyPairGenerator", "RSA//PKCS1-v1-5-SHA-256",
				"de.cardcontact.smartcardhsmprovider.KeyPairGenerator$RSA"));

		putService(new KeyPairGenService(this, "KeyPairGenerator", "RSA//PKCS1-PSS-SHA-1",
				"de.cardcontact.smartcardhsmprovider.KeyPairGenerator$RSA"));

		putService(new KeyPairGenService(this, "KeyPairGenerator", "RSA//PKCS1-PSS-SHA-256",
				"de.cardcontact.smartcardhsmprovider.KeyPairGenerator$RSA"));

		alias = new Vector<String>();
		alias.addElement("EC");
		putService(new KeyPairGenService(this, "KeyPairGenerator", "EC//ECDSA-SHA-1",
				"de.cardcontact.smartcardhsmprovider.KeyPairGenerator$EC", alias));

		putService(new KeyPairGenService(this, "KeyPairGenerator", "EC//ECDSA-SHA-224",
				"de.cardcontact.smartcardhsmprovider.KeyPairGenerator$EC"));

		putService(new KeyPairGenService(this, "KeyPairGenerator", "EC//ECDSA-SHA-256",
				"de.cardcontact.smartcardhsmprovider.KeyPairGenerator$EC"));

		put("AlgorithmParameters.EC SupportedCurves", "secp192r1,prime192v1,secp256r1,prime256v1,brainpoolP192r1,brainpoolP224r1,brainpoolP256r1,brainpoolP320r1,secp192k1,secp256k1");

		// Signature

		alias = new Vector<String>();
		alias.addElement("SHA1withRSA");
		putService(new SCHSMService(this, "Signature", "SHA1withRSA//PKCS1-v1-5",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMSignature$SHA1withRSAPKCS1V15", alias));

		alias = new Vector<String>();
		alias.addElement("SHA256withRSA");
		putService(new SCHSMService(this, "Signature", "SHA256withRSA//PKCS1-v1-5",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMSignature$SHA256withRSAPKCS1V15", alias));

		alias = new Vector<String>();
		alias.addElement("SHA1withRSA/PSS");
		putService(new SCHSMService(this, "Signature", "SHA1withRSA//PKCS1-PSS",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMSignature$SHA1withRSAPKCS1PSS", alias));

		alias = new Vector<String>();
		alias.addElement("SHA256withRSA/PSS");
		putService(new SCHSMService(this, "Signature", "SHA256withRSA//PKCS1-PSS",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMSignature$SHA256withRSAPKCS1PSS", alias));

		putService(new SCHSMService(this, "Signature", "SHA1withECDSA",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMSignature$SHA1withECDSA"));

		putService(new SCHSMService(this, "Signature", "SHA224withECDSA",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMSignature$SHA224withECDSA"));

		putService(new SCHSMService(this, "Signature", "SHA256withECDSA",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMSignature$SHA256withECDSA"));

		putService(new SCHSMService(this, "Signature", "NONEwithECDSA",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMSignature$NONEwithECDSA"));

		putService(new SCHSMService(this, "Signature", "NONEwithRSA",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMSignature$NONEwithRSA"));

		// KeyStore

		putService(new SCHSMService(this, "KeyStore", "SmartCardHSMKeyStore",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMKeyStore"));

		// Cipher
		alias = new Vector<String>();
		alias.addElement("RSA");
		putService(new SCHSMService(this, "Cipher", "RSA/None/NoPadding",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMCipher", alias));

		alias = new Vector<String>();
		alias.addElement("RSA/ECB/PKCS1Padding");
		alias.addElement("RSA/NONE/PKCS1Padding");
		putService(new SCHSMService(this, "Cipher", "RSA/None/PKCS1Padding",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMCipher", alias));

		alias = new Vector<String>();
		alias.addElement("RSA/NONE/OAEPWithSHA1AndMGF1Padding");
		alias.addElement("RSA/None/OAEPWithSHA1AndMGF1Padding");
		putService(new SCHSMService(this, "Cipher", "RSA/ECB/OAEPWithSHA-1AndMGF1Padding",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMCipher", alias));

		// ECDH
		putService(new SCHSMService(this, "KeyAgreement", "ECDH",
				"de.cardcontact.smartcardhsmprovider.SmartCardHSMKeyAgreement"));

	}



	public void setSmartCardHSMCardService(SmartCardHSMCardService schsm) {
		if (schsm == null) {
			this.schsm = null;
			return;
		}
		// Get card object
		try {
			log.debug("Setting SmartCardHSMCardService object");
			schsm.getCard().setAPDUTracer(this);

			schsm.useClassThreePinPad(usePinPad);

			if (doSecureMessaging) {
				schsm.initSecureMessaging();
			}

		} catch (OpenCardException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}

		this.schsm = schsm;
	}



	/**
	 * Returns an object to get access to the card
	 *
	 * @return SmartCardHSMCardService
	 */
	public SmartCardHSMCardService getSmartCardHSMCardService() {

		checkCardState();

		return schsm;
	}

	/**
	 * Check whether a card object is initialized.
	 * If not then try to create an card object
	 * or throw a ProviderException if no card is inserted.
	 * @throws ProviderException if no card is inserted
	 */
	private void checkCardState() {

		if (schsm == null) {

			try {
				EventGenerator.getGenerator().createEventsForPresentCards(this);
			} catch (CardTerminalException e) {
				log.debug(e.getLocalizedMessage(), e);
			}
			if (schsm == null) {
				throw new ProviderException("No card inserted");
			}
		}
	}



	/**
	 * Verify the card's pin.
	 * The pin will be obtained from a callback mechanism
	 * or will be entered on a class 3 card reader
	 * depending on the settings made by usePinPad
	 *
	 * @param cbh The CallbackHandler used by this provider to obtain authentication information from the caller
	 * @param subject This parameter is not used, set it to null
	 */
	@Override
	public void login(Subject subject, CallbackHandler cbh) throws LoginException {

		checkCardState();

		/*
		 * No callback handler and no class 3 reader
		 */
		if (this.callBackHandler == null && cbh == null && usePinPad == false) {
			throw new LoginException("The mandatory CallbackHandler cannot be null");
		}

		if (cbh != null) {
			setCallbackHandler(cbh);
		}

		boolean verified = verify();

		if (!verified) {
			log.info("Login failed. Wrong PIN?");
			throw new FailedLoginException("Login failed. Wrong PIN?");
		}
	}

	@Override
	public void logout() throws LoginException {
	}

	/**
	 * Set the CallbackHandler.
	 * The provider will immediately use this handler
	 * to query the password.
	 *
	 * @param cbh The CallbackHandler used by this provider to obtain authentication information from the caller
	 */
	@Override
	public void setCallbackHandler(CallbackHandler cbh) {
		this.callBackHandler = cbh;
	}

	/**
	 * Start OCF and register to an event generator.
	 * Then create an event to instantiate the card if present.
	 *
	 * @throws OpenCardPropertyLoadingException
	 * @throws CardServiceException
	 * @throws CardTerminalException
	 * @throws ClassNotFoundException
	 */
	public void register() throws OpenCardPropertyLoadingException, CardServiceException, CardTerminalException, ClassNotFoundException {
		if (SmartCard.isStarted()) {
			log.debug("OCF already running");
			if (doOCFShutdown == 0) {				// Started outside of the provider
				doOCFShutdown++;
			}
		} else {
			log.debug("Startup of OCF...");
			SmartCard.startup();

			CardTerminalRegistry ctr = CardTerminalRegistry.getRegistry();
			CardTerminalFactory ctf = new SmartCardIOFactory();
			//			String param[] = { "*", "PCSC-NOPOLL" };
			String param[] = { "*", "PCSC" };
			ctf.createCardTerminals(ctr, param);

			CardServiceRegistry csr = CardServiceRegistry.getRegistry();
			CardServiceFactory csf = new IsoCardServiceFactory();

			csr.add(csf);
			csf = new PassThruCardServiceFactory();
			csr.add(csf);
			csf = new GlobalPlatformCardServiceFactory();
			csr.add(csf);
			csf = new SmartCardHSMCardServiceFactory();
			csr.add(csf);
			csf = new RemoteClientCardServiceFactory();
			csr.add(csf);
		}
		doOCFShutdown++;

		log.debug("Register to event generator...");
		EventGenerator.getGenerator().addCTListener(this);
	}

	/**
	 * Unregister from event generator and shutdown.
	 */
	public void unregister() {
		log.debug("Unregister from event generator...");
		EventGenerator.getGenerator().removeCTListener(this);
		doOCFShutdown--;
		if (doOCFShutdown <= 0) {
			try {
				log.debug("Shutdown OCF...");
				SmartCard.shutdown();
			} catch (CardTerminalException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}
		}
		sc = null;
		ct = null;
		schsm = null;
	}

	@Override
	public void cardInserted(CardTerminalEvent ctEvent)
			throws CardTerminalException {
		if (sc == null) {
			log.debug("Card inserted");
			CardTerminalRegistry ctr = CardTerminalRegistry.getRegistry();

			if (ct == null) {
				if (terminalName != null) {
					ct = ctr.cardTerminalForName(terminalName);
				}

				if (ct == null) {
					ct = ctEvent.getCardTerminal();
				}
			}
			CardRequest cr = new CardRequest(CardRequest.ANYCARD, ct, SmartCardHSMCardService.class);
			sc = SmartCard.getSmartCard(ctEvent, cr);
			if (sc == null) {
				log.debug("New card is not a SmartCard-HSM");
				return;
			}
			slotID = ctEvent.getSlotID();

			// Get card object
			try {
				log.debug("New SmartCardHSMCardService object");
				schsm = (SmartCardHSMCardService) sc.getCardService(SmartCardHSMCardService.class, true);

				schsm.getCard().setAPDUTracer(this);

				schsm.useClassThreePinPad(usePinPad);

				if (doSecureMessaging) {
					schsm.initSecureMessaging();
				}

			} catch (CardServiceException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			} catch (ClassNotFoundException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}

			/*
			 * If the provider is logged
			 * then verify the card's pin
			 */
			if (callBackHandler != null) {
				verify();
			}
		}

	}

	@Override
	public void cardRemoved(CardTerminalEvent ctEvent)
			throws CardTerminalException {
		if (ctEvent.getSlotID() == slotID) {
			if (ct != null && !ctEvent.getCardTerminal().equals(ct)) {
				return;
			}
			log.debug("Card removed");
			sc = null;
			ct = null;
			slotID = -1;
			schsm = null;
		}
	}

	/**
	 * This class instantiates the service provider classes
	 */
	private static class SCHSMService extends Service {

		/**
		 * The parameter types of the constructor
		 */
		private static final Class[] constructorParamTypes = { SmartCardHSMProvider.class, String.class};

		/**
		 *
		 * @param provider
		 * @param type
		 * @param algorithm
		 * @param className
		 * @param aliases
		 */
		SCHSMService(AuthProvider provider, String type, String algorithm, String className, Vector<String> aliases) {
			super(provider, type, algorithm, className, aliases, null);
		}

		/**
		 *
		 * @param provider
		 * @param type
		 * @param algorithm
		 * @param className
		 */
		SCHSMService(AuthProvider provider, String type, String algorithm, String className) {
			super(provider, type, algorithm, className, null, null);
		}

		public Object newInstance(Object param) {
			try	{

				Class clazz;
				SmartCardHSMProvider provider = (SmartCardHSMProvider) getProvider();
				ClassLoader loader = provider.getClass().getClassLoader();
				if (loader == null) {
					clazz = Class.forName(getClassName());
				} else {
					clazz = loader.loadClass(getClassName());
				}

				Constructor cons = clazz.getConstructor(constructorParamTypes);

				Object obj = cons.newInstance(new Object[] {provider, getAlgorithm()});
				return obj;

			} catch (Exception e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}
		}
	}



	/**
	 * This class instantiates the KeyPairGenerator service provider
	 */
	private static class KeyPairGenService extends Service {

		/**
		 * The parameter types of the constructor
		 */
		private static final Class[] constructorParamTypes = { SmartCardHSMProvider.class, String.class};

		/**
		 *
		 * @param provider
		 * @param type
		 * @param algorithm
		 * @param className
		 * @param aliases
		 */
		KeyPairGenService(AuthProvider provider, String type, String algorithm, String className, Vector<String> aliases) {
			super(provider, type, algorithm, className, aliases, null);
		}

		/**
		 *
		 * @param provider
		 * @param type
		 * @param algorithm
		 * @param className
		 */
		KeyPairGenService(AuthProvider provider, String type, String algorithm, String className) {
			super(provider, type, algorithm, className, null, null);
		}

		public Object newInstance(Object param) {
			try	{

				Class clazz;
				SmartCardHSMProvider provider = (SmartCardHSMProvider) getProvider();
				ClassLoader loader = provider.getClass().getClassLoader();
				if (loader == null) {
					clazz = Class.forName(getClassName());
				} else {
					clazz = loader.loadClass(getClassName());
				}

				Constructor cons = clazz.getConstructor(constructorParamTypes);

				StringTokenizer st = new StringTokenizer(getAlgorithm(), "/");
				st.nextToken();

				// alternative
				//String padding = getAlgorithm().substring(getAlgorithm().lastIndexOf('/') + 1);

				Object obj = cons.newInstance(new Object[] {provider, st.nextToken()});
				return obj;

			} catch (Exception e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}
		}
	}



	@Override
	public void traceAnswerToReset(SlotChannel sc, CardID cardID) {

		String s = sc.getCardTerminal().getName();
		s = s.concat("\n");
		s = s.concat(HexString.hexifyByteArray(cardID.getATR()));

		log.debug(s);
	}



	@Override
	public void traceCommandAPDU(SlotChannel sc, CommandAPDU capdu) {

		String s = sc.getCardTerminal().getName();
		s = s.concat("\n");
		s = s.concat(APDUFormatter.commandAPDUToString(capdu));

		log.debug(s);
	}



	@Override
	public void traceResponseAPDU(SlotChannel sc, ResponseAPDU rapdu) {
		try {
			StringBuffer sb = new StringBuffer(80);
			int len = rapdu.getLength();
			byte[] buf = rapdu.getBuffer();

			sb.append("   R: ");
			sb.append(StatusWordTable.MessageForSW(rapdu.sw()));
			sb.append(" Lr=" + (len - 2));
			sb.append("\n");

			if (len > 2) {
				sb.append(HexString.dump(buf, 0, len - 2, 16, 6));
			}

			log.debug(sb.toString());
		}
		catch(Exception e) {
			log.debug("Error decoding APDU.", e);
		}
	}
}
