/*
 *  ---------
 * |.##> <##.|  Open Smart Card Development Platform (www.openscdp.org)
 * |#       #|
 * |#       #|  Copyright (c) 1999-2016 CardContact Systems GmbH
 * |'##> <##'|  32429 Minden, Germany (www.cardcontact.de)
 *  ---------
 *
 *  This file is part of OpenSCDP.
 *
 *  OpenSCDP is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  OpenSCDP is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with OpenSCDP; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.cardcontact.smartcardhsmprovider;

import java.io.ByteArrayOutputStream;
import java.security.InvalidKeyException;
import java.security.InvalidParameterException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.ProviderException;
import java.security.PublicKey;
import java.security.SignatureException;
import java.security.SignatureSpi;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMCardService;
import de.cardcontact.opencard.service.smartcardhsm.SmartCardHSMPrivateKey;
import opencard.core.service.CardServiceException;
import opencard.core.terminal.CardTerminalException;

/**
 * This class provides the functionality of a digital signature algorithm.
 *
 * @author lew
 */
public class SmartCardHSMSignature extends SignatureSpi {

	private final static Logger log = LoggerFactory.getLogger(SmartCardHSMSignature.class);

	/**
	 * For signing this state has the value true, for verification the value
	 * false.
	 */
	private static boolean SIGN_STATE;

	/**
	 * This object provides access to the SmartCard.
	 */
	protected SmartCardHSMCardService schsm;

	/**
	 * The key reference to the private key on the card.
	 */
	protected SmartCardHSMPrivateKey privateKey;

	/**
	 * The algorithm used to create the message digest.
	 */
	private String digestAlgorithm;

	/**
	 * The algorithm for the digital signature.
	 */
	protected String signAlgorithm;

	/**
	 * The algorithm used to pad the message digest. Must be one of PKCS1_V15,
	 * PKCS1_PSS or NONE.
	 */
	private String padAlgorithm;

	/**
	 * This object creates the message digest.
	 */
	private MessageDigest digest;

	public SmartCardHSMSignature(SmartCardHSMProvider provider, String digestAlgorithm, String signAlgorithm, String padAlgorithm) {

		this.schsm = provider.getSmartCardHSMCardService();
		this.digestAlgorithm = digestAlgorithm;
		this.signAlgorithm = signAlgorithm;
		this.padAlgorithm = padAlgorithm;

		if (!provider.isVerified()) throw new ProviderException("Login required.");
	}

	@Override
	protected Object engineGetParameter(String arg0) throws InvalidParameterException {
		return null;
	}

	@Override
	protected void engineInitSign(PrivateKey key) throws InvalidKeyException {
		log.trace("SmartCardHSMSignature.engineInitSign");
		SIGN_STATE = true;

		if (key instanceof SmartCardHSMPrivateKey) {
			this.privateKey = (SmartCardHSMPrivateKey) key;
		} else {
			throw new InvalidKeyException("The key is not type of SmartCardHSMPrivateKey");
		}

		try {
			digest = MessageDigest.getInstance(digestAlgorithm);
		} catch (NoSuchAlgorithmException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}
	}

	@Override
	protected void engineInitVerify(PublicKey arg0) throws InvalidKeyException {
		SIGN_STATE = false;
		log.debug("Do nothing...");
	}

	@Override
	protected void engineSetParameter(String arg0, Object arg1) throws InvalidParameterException {

	}

	/**
	 * @return the signature bytes
	 */
	@Override
	protected byte[] engineSign() throws SignatureException {
		byte[] md = digest.digest();
		byte[] signature = null;
		try {
			signature = schsm.signHash(this.privateKey, signAlgorithm, padAlgorithm, md);
		} catch (CardServiceException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		} catch (CardTerminalException e) {
			log.error(e.getLocalizedMessage(), e);
			throw new ProviderException(e);
		}

		return signature;
	}

	@Override
	protected void engineUpdate(byte arg0) throws SignatureException {
		log.trace("SmartCardHSMSignature.engineUpdate()");
		if (SIGN_STATE) {
			digest.update(arg0);
		}
	}

	@Override
	protected void engineUpdate(byte[] arg0, int arg1, int arg2) throws SignatureException {
		log.trace("SmartCardHSMSignature.engineUpdate()");
		if (SIGN_STATE) {
			digest.update(arg0, arg1, arg2);
		}
	}

	@Override
	protected boolean engineVerify(byte[] arg0) throws SignatureException {
		log.trace("SmartCardHSMSignature.engineVerify");
		return false;
	}

	/**
	 * Get a digital signature by using the signature algorithm SHA1withRSA and
	 * the pad algorithm PKCS1-v15.
	 *
	 * The Data will be hashed with SHA1.
	 *
	 */
	protected static class SHA1withRSAPKCS1V15 extends SmartCardHSMSignature {
		public SHA1withRSAPKCS1V15(SmartCardHSMProvider provider, String algorithm) {
			super(provider, "SHA1", "SHA1withRSA", "PKCS1_V15");
		}
	}

	/**
	 * Get a digital signature by using the signature algorithm SHA256withRSA
	 * and the pad algorithm PKCS1-v15.
	 *
	 * The Data will be hashed with SHA256.
	 *
	 */
	protected static class SHA256withRSAPKCS1V15 extends SmartCardHSMSignature {
		public SHA256withRSAPKCS1V15(SmartCardHSMProvider provider, String algorithm) {
			super(provider, "SHA256", "SHA256withRSA", "PKCS1_V15");
		}
	}

	/**
	 * Get a digital signature by using the signature algorithm SHA256withRSA
	 * and the pad algorithm PKCS1 EMSA-PSS.
	 *
	 * The Data will be hashed with SHA256.
	 *
	 */
	protected static class SHA256withRSAPKCS1PSS extends SmartCardHSMSignature {
		public SHA256withRSAPKCS1PSS(SmartCardHSMProvider provider, String algorithm) {
			super(provider, "SHA256", "SHA256withRSA", "PKCS1_PSS");
		}
	}

	/**
	 * Get a digital signature by using the signature algorithm SHA1withRSA
	 * and the pad algorithm PKCS1 EMSA-PSS.
	 *
	 * The Data will be hashed with SHA1.
	 *
	 */
	protected static class SHA1withRSAPKCS1PSS extends SmartCardHSMSignature {
		public SHA1withRSAPKCS1PSS(SmartCardHSMProvider provider, String algorithm) {
			super(provider, "SHA1", "SHA1withRSA", "PKCS1_PSS");
		}
	}

	/**
	 * Inner class for signing data that was hashed by the application.
	 *
	 */
	protected static class NONEwithRSA extends SmartCardHSMSignature {
		private ByteArrayOutputStream md;

		public NONEwithRSA(SmartCardHSMProvider provider, String algorithm) {
			super(provider, "", "NONEwithRSA", "PKCS1_V15");
			log.debug("NONEwithRSA");
		}

		@Override
		protected void engineUpdate(byte arg0) throws SignatureException {
			if (SIGN_STATE) {
				md.write(arg0);
			}
		}

		@Override
		protected void engineUpdate(byte[] arg0, int arg1, int arg2) throws SignatureException {
			if (SIGN_STATE) {
				md.write(arg0, arg1, arg2);
			}
		}

		@Override
		protected void engineInitSign(PrivateKey key) throws InvalidKeyException {
			log.debug("Initialized for signature...");
			SIGN_STATE = true;
			this.privateKey = (SmartCardHSMPrivateKey) key;
			md = new ByteArrayOutputStream();
		}

		/**
		 * @return the signature bytes or null
		 */
		@Override
		protected byte[] engineSign() throws SignatureException {
			byte[] signature = null;

			try {
				signature = schsm.signHash(this.privateKey, signAlgorithm, md.toByteArray());
			} catch (CardServiceException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			} catch (CardTerminalException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			} catch (InvalidKeyException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}

			return signature;
		}
	}


	/**
	 * Inner class for signing data that was hashed by the application.
	 *
	 */
	protected static class NONEwithECDSA extends SmartCardHSMSignature {
		private ByteArrayOutputStream md;

		public NONEwithECDSA(SmartCardHSMProvider provider, String algorithm) {
			super(provider, "", "NONEwithECDSA", "NONE");
			log.debug("NONEwithECDSA");
		}

		@Override
		protected void engineUpdate(byte arg0) throws SignatureException {
			if (SIGN_STATE) {
				md.write(arg0);
			}
		}

		@Override
		protected void engineUpdate(byte[] arg0, int arg1, int arg2) throws SignatureException {
			if (SIGN_STATE) {
				md.write(arg0, arg1, arg2);
			}
		}

		@Override
		protected void engineInitSign(PrivateKey key) throws InvalidKeyException {
			log.debug("Initialized for signature...");
			SIGN_STATE = true;
			this.privateKey = (SmartCardHSMPrivateKey) key;
			md = new ByteArrayOutputStream();
		}

		/**
		 * @return the signature bytes or null
		 */
		@Override
		protected byte[] engineSign() throws SignatureException {
			byte[] signature = null;

			try {
				signature = schsm.signHash(this.privateKey, signAlgorithm, md.toByteArray());
			} catch (CardServiceException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			} catch (CardTerminalException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			} catch (InvalidKeyException e) {
				log.error(e.getLocalizedMessage(), e);
				throw new ProviderException(e);
			}

			return signature;
		}

	}


	/**
	 * Get a digital signature by using the signature algorithm SHA1withECDSA
	 *
	 * The Data will be hashed with SHA1.
	 *
	 */
	protected static class SHA1withECDSA extends SmartCardHSMSignature {
		public SHA1withECDSA(SmartCardHSMProvider provider, String algorithm) {
			super(provider, "SHA1", "NONEwithECDSA", "NONE");
		}
	}

	/**
	 * Get a digital signature by using the signature algorithm SHA224withECDSA
	 *
	 * The Data will be hashed with SHA224.
	 *
	 */
	protected static class SHA224withECDSA extends SmartCardHSMSignature {
		public SHA224withECDSA(SmartCardHSMProvider provider, String algorithm) {
			super(provider, "SHA-224", "NONEwithECDSA", "NONE");
		}
	}

	/**
	 * Get a digital signature by using the signature algorithm SHA256withECDSA
	 *
	 * The Data will be hashed with SHA256.
	 *
	 */
	protected static class SHA256withECDSA extends SmartCardHSMSignature {
		public SHA256withECDSA(SmartCardHSMProvider provider, String algorithm) {
			super(provider, "SHA-256", "NONEwithECDSA", "NONE");
		}
	}
}
